""" Functions for writing and loading backup files """

import datetime
import json
import logging
import os
from enum import Enum

from osxmetadata import OSXMetaData, __version__

__all__ = ["get_backup_dict", "write_backup_file", "load_backup_file"]


class BackupDatabaseType(Enum):
    SINGLE_RECORD_JSON = 1
    JSON = 2


def get_backup_dict(filepath: str):
    """get backup dict for a single file"""
    md = OSXMetaData(filepath)
    try:
        backup_dict = md.asdict()
    except Exception as e:
        logging.warning(f"Error retrieving metadata for file {filepath}: {e}")
        backup_dict = {}

    backup_dict.update(
        {
            "_version": __version__,
            "_filename": md._fname.name,
            "_filepath": md._posix_path,
        }
    )
    return backup_dict


def backup_database_type(path: str) -> BackupDatabaseType:
    """Return BackupDatabaseType enum indicating type of backup file for path"""
    with open(path) as fp:
        line = fp.readline()
        if not line:
            raise ValueError("Unknown backup file type")
        ch = line[0]
        if ch == "{":
            return BackupDatabaseType.SINGLE_RECORD_JSON
        elif ch == "[":
            return BackupDatabaseType.JSON
        else:
            raise ValueError("Unknown backup file type")


def write_backup_file(backup_file, backup_data):
    """Write backup_data to backup_file as JSON
    backup_data: dict where key is filename and value is dict of the attributes
    as returned by json.loads(OSXMetaData.to_json())"""

    # Convert datetime objects to isoformat strings for serialization
    for filename, data in backup_data.items():
        for key, value in data.items():
            if isinstance(value, datetime.datetime):
                data[key] = value.isoformat()
            elif isinstance(value, (list, tuple)):
                if value and isinstance(value[0], datetime.datetime):
                    data[key] = [v.isoformat() for v in value]
        # strip null values
        data = {k: v for k, v in data.items() if v is not None}
        backup_data[filename] = data

    with open(backup_file, mode="w") as fp:
        json.dump(list(backup_data.values()), fp, indent=2)


def load_backup_file(backup_file):
    """Load attribute data from JSON in backup_file
    Returns: backup_data dict"""

    if not os.path.isfile(backup_file):
        raise FileNotFoundError(f"Could not find backup file: {backup_file}")

    if backup_database_type(backup_file) == BackupDatabaseType.SINGLE_RECORD_JSON:
        # old style single record of json per file
        backup_data = {}
        with open(backup_file, mode="r") as fp:
            for line in fp:
                data = json.loads(line)
                fname = data["_filename"]
                if fname in backup_data:
                    logging.warning(
                        f"WARNING: duplicate filename {fname} found in {backup_file}"
                    )

                backup_data[fname] = data
    else:
        with open(backup_file, mode="r") as fp:
            backup_records = json.load(fp)
        backup_data = {data["_filename"]: data for data in backup_records}

    return backup_data



"""Test osxmetadata basic mditem metadata tags"""

import os

import pytest

from osxmetadata import OSXMetaData
from osxmetadata.attribute_data import (
    MDITEM_ATTRIBUTE_AUDIO,
    MDITEM_ATTRIBUTE_DATA,
    MDITEM_ATTRIBUTE_IMAGE,
    MDITEM_ATTRIBUTE_READ_ONLY,
    MDITEM_ATTRIBUTE_VIDEO,
)

from .conftest import FINDER_COMMENT_SNOOZE, snooze, value_for_type

# filter out the read-only attributes
MDITEM_ATTRIBUTES_TO_TEST = [
    a["name"]
    for a in MDITEM_ATTRIBUTE_DATA.values()
    if a["name"] not in MDITEM_ATTRIBUTE_READ_ONLY
    and a["name"] not in MDITEM_ATTRIBUTE_AUDIO
    and a["name"] not in MDITEM_ATTRIBUTE_IMAGE
    and a["name"] not in MDITEM_ATTRIBUTE_VIDEO
]

# Not all attributes can be cleared by setting to None
MDITEM_ATTRIBUTES_CAN_BE_REMOVED = [
    a for a in MDITEM_ATTRIBUTES_TO_TEST if a not in ["kMDItemContentModificationDate"]
]


@pytest.mark.parametrize("attribute_name", MDITEM_ATTRIBUTES_TO_TEST)
def test_mditem_attributes_get_set(attribute_name, test_file):
    """test mditem attributes"""

    # can't use tmp_path fixture because the tmpfs filesystem doesn't support xattrs
    attribute = MDITEM_ATTRIBUTE_DATA[attribute_name]
    attribute_type = attribute["python_type"]
    test_value = value_for_type(attribute_type)

    md = OSXMetaData(test_file.name)
    md.set(attribute_name, test_value)
    snooze()
    if attribute_name == "kMDItemFinderComment":
        snooze(FINDER_COMMENT_SNOOZE)
    assert md.get(attribute_name) == test_value


@pytest.mark.parametrize("attribute_name", MDITEM_ATTRIBUTES_TO_TEST)
def test_mditem_attributes_dict(attribute_name, test_file):
    """test mditem attributes with dict access"""

    attribute = MDITEM_ATTRIBUTE_DATA[attribute_name]
    attribute_type = attribute["python_type"]
    test_value = value_for_type(attribute_type)

    md = OSXMetaData(test_file.name)
    md[attribute_name] = test_value
    snooze()
    if attribute_name == "kMDItemFinderComment":
        snooze(FINDER_COMMENT_SNOOZE)
    assert md[attribute_name] == test_value


@pytest.mark.parametrize("attribute_name", MDITEM_ATTRIBUTES_TO_TEST)
def test_mditem_attributes_property(attribute_name, test_file):
    """test mditem attributes with property access"""

    attribute = MDITEM_ATTRIBUTE_DATA[attribute_name]
    attribute_type = attribute["python_type"]
    test_value = value_for_type(attribute_type)

    md = OSXMetaData(test_file.name)
    setattr(md, attribute_name, test_value)
    snooze()
    if attribute_name == "kMDItemFinderComment":
        snooze(FINDER_COMMENT_SNOOZE)
    assert getattr(md, attribute_name) == test_value


@pytest.mark.parametrize("attribute_name", MDITEM_ATTRIBUTES_TO_TEST)
def test_mditem_attributes_short_name(attribute_name, test_file):
    """test mditem attributes"""

    attribute = MDITEM_ATTRIBUTE_DATA[attribute_name]
    attribute_type = attribute["python_type"]
    test_value = value_for_type(attribute_type)

    md = OSXMetaData(test_file.name)
    setattr(md, attribute["short_name"], test_value)
    snooze()
    if attribute_name == "kMDItemFinderComment":
        snooze(FINDER_COMMENT_SNOOZE)
    assert getattr(md, attribute["short_name"]) == test_value


@pytest.mark.parametrize("attribute_name", MDITEM_ATTRIBUTE_DATA.keys())
def test_mditem_attributes_all(attribute_name, test_file):
    """Test that all attributes can be accessed without error"""

    md = OSXMetaData(test_file.name)
    md.get(attribute_name)


# this test failes on kMDItemFinderComment though the code works when run outside pytest
@pytest.mark.parametrize(
    "attribute_name",
    [
        attr
        for attr in MDITEM_ATTRIBUTES_CAN_BE_REMOVED
        if attr != "kMDItemFinderComment"
    ],
)
def test_mditem_attributes_set_none(attribute_name, test_file):
    """test mditem attributes can be set to None to remove"""

    # can't use tmp_path fixture because the tmpfs filesystem doesn't support xattrs
    attribute = MDITEM_ATTRIBUTE_DATA[attribute_name]
    attribute_type = attribute["python_type"]
    test_value = value_for_type(attribute_type)
    md = OSXMetaData(test_file.name)
    md.set(attribute_name, test_value)
    snooze()
    if attribute_name == "kMDItemFinderComment":
        snooze(FINDER_COMMENT_SNOOZE)  # Finder needs a moment to update the comment
    assert md.get(attribute_name)
    md.set(attribute_name, None)
    snooze()
    if attribute_name == "kMDItemFinderComment":
        snooze(FINDER_COMMENT_SNOOZE)
    assert not md.get(attribute_name)


@pytest.mark.skipif(
    bool(os.environ.get("GITHUB_ACTION")), reason="GitHub Actions doesn't run md import"
)
def test_mditem_attributes_image(test_image):
    """test mditem attributes for image files"""

    md = OSXMetaData(test_image)
    assert md.get("kMDItemLatitude") == "-34.91889166666667"
    assert md.get("kMDItemPixelHeight") == 2754


@pytest.mark.skipif(
    bool(os.environ.get("GITHUB_ACTION")), reason="GitHub Actions doesn't run md import"
)
def test_mditem_attributes_video(test_video):
    """test mditem attributes for video files"""

    md = OSXMetaData(test_video)
    assert "H.264" in md.get("kMDItemCodecs")
    assert md.get("kMDItemAudioBitRate") == 64.0


@pytest.mark.skipif(
    bool(os.environ.get("GITHUB_ACTION")), reason="GitHub Actions doesn't run md import"
)
def test_mditem_attributes_audio(test_audio):
    """test mditem attributes for audio files"""

    md = OSXMetaData(test_audio)
    assert md.get("kMDItemAudioSampleRate") == 44100.0


def test_get_set_mditem_attribute_value(test_file):
    """test get and set of mditem attribute value using the direct methods without value conversion, #83"""

    md = OSXMetaData(test_file.name)
    md.set_mditem_attribute_value("kMDItemComment", "foo,bar")
    snooze()
    assert md.get_mditem_attribute_value("kMDItemComment") == "foo,bar"
    assert md.comment == "foo,bar"


def test_attribute_get_set(test_file):
    """Test direct access get/set attribute values"""

    md = OSXMetaData(test_file.name)
    assert not md.authors
    md.authors = ["foo", "bar"]
    snooze()
    assert md.authors == ["foo", "bar"]
    md.authors = ["bar"]
    snooze()
    assert md.authors == ["bar"]
    md.set("authors", ["foo"])
    snooze()
    assert md.authors == ["foo"]



"""Test osxmetadata get/set/remove xattr methods"""

import plistlib

import pytest

from osxmetadata import OSXMetaData

from .conftest import snooze


def test_xattr_get_set_remove(test_file):
    """test get/set/remove xattr methods"""

    attribute = "com.apple.metadata:kMDItemComment"
    value = "This is my comment"

    md = OSXMetaData(test_file.name)
    md.comment = value
    snooze()
    xattr_comment = md.get_xattr(attribute, decode=plistlib.loads)
    assert xattr_comment == value

    value = "This is my new comment"
    md.set_xattr(attribute, value, encode=plistlib.dumps)
    snooze()
    xattr_comment = md.get_xattr(attribute, decode=plistlib.loads)
    assert xattr_comment == value

    md.remove_xattr(attribute)
    with pytest.raises(KeyError):
        md.get_xattr(attribute, decode=plistlib.loads)



"""Test osxmetadata path property"""

import os

from osxmetadata import OSXMetaData


def test_asdict(test_file):
    """Test asdict method"""
    md = OSXMetaData(test_file.name)
    cwd = os.getcwd()
    assert md.path == os.path.join(cwd, test_file.name)



"""Test findercomment """

import pytest

from osxmetadata import OSXMetaData, kMDItemFinderComment

from .conftest import FINDER_COMMENT_SNOOZE, snooze


@pytest.mark.skip(
    "This should pass but on my machine (Catalina 10.15.7) it does not; the code runs correctly outside of pytest"
)
def test_finder_comments(test_file):

    md = OSXMetaData(test_file.name)
    fc = "This is my new comment"
    md.findercomment = fc
    # Finder comment is set via AppleScript events and may take a moment to update
    snooze(FINDER_COMMENT_SNOOZE)
    assert md.findercomment == fc

    fc += ", foo"
    md.findercomment += ", foo"
    snooze(FINDER_COMMENT_SNOOZE)
    assert md.findercomment == fc

    # set finder comment to "" deletes it as this mirrors Finder and mdls
    md.findercomment = ""
    snooze(FINDER_COMMENT_SNOOZE)
    assert not md.findercomment

    md.findercomment = "foo"
    snooze(FINDER_COMMENT_SNOOZE)
    assert md.findercomment == "foo"

    # set finder comment to None deletes it
    md.findercomment = None
    snooze(FINDER_COMMENT_SNOOZE)
    assert not md.findercomment

    # can we set findercomment after is was set to None?
    md.findercomment = "bar"
    snooze(FINDER_COMMENT_SNOOZE)
    assert md.findercomment == "bar"


@pytest.mark.skip(
    "This should pass but on my machine (Catalina 10.15.7) it does not; the code runs correctly outside of pytest"
)
def test_finder_comments_get_set(test_file):
    """test get/set attribute"""

    attribute = kMDItemFinderComment

    md = OSXMetaData(test_file.name)
    fc = "This is my new comment"
    md.set(attribute, fc)
    snooze(FINDER_COMMENT_SNOOZE)
    assert md.findercomment == fc

    fc += ", foo"
    md.findercomment += ", foo"
    snooze(FINDER_COMMENT_SNOOZE)
    assert md.findercomment == fc

    # set finder comment to "" deletes it as this mirrors mdls and Finder
    md.set(attribute, "")
    snooze(FINDER_COMMENT_SNOOZE)
    assert not md.get(attribute)

    md.set(attribute, "foo")
    snooze(FINDER_COMMENT_SNOOZE)
    assert md.get(attribute) == "foo"

    # set finder comment to None deletes it
    md.set(attribute, None)
    snooze(FINDER_COMMENT_SNOOZE)
    assert not md.get(attribute)

    # can we set findercomment after is was set to None?
    md.set(attribute, "bar")
    snooze(FINDER_COMMENT_SNOOZE)
    assert md.get(attribute) == "bar"


def test_finder_comments_dir(test_dir):
    """test get/set attribute but on a directory, not on a file"""

    attribute = kMDItemFinderComment

    md = OSXMetaData(test_dir)
    fc = "This is my new comment"
    md.set(attribute, fc)

    snooze(FINDER_COMMENT_SNOOZE)
    assert md.findercomment == fc

    md.findercomment += ", foo"
    fc += ", foo"

    snooze(FINDER_COMMENT_SNOOZE)
    assert md.findercomment == fc



"""Config for pytest"""

import datetime
import os
import time
import typing as t
from tempfile import NamedTemporaryFile, TemporaryDirectory

import pytest

TEST_IMAGE = "tests/test_image.jpg"
TEST_VIDEO = "tests/test_video.mov"
TEST_AUDIO = "tests/test_audio.m4a"

# how long to wait for metadata to be written to disk
# if running in GitHub Actions, wait longer
GH_ACTION_SNOOZE = 10.0
SNOOZE_TIME = GH_ACTION_SNOOZE if os.environ.get("GITHUB_ACTION") else 0.1
# some tests need a longer snooze time
LONG_SNOOZE = GH_ACTION_SNOOZE if os.environ.get("GITHUB_ACTION") else 3.0
# Finder comments need more time to be written to disk
FINDER_COMMENT_SNOOZE = LONG_SNOOZE

def snooze(seconds: float = SNOOZE_TIME) -> None:
    """Sleep for a bit to allow Finder to update metadata"""
    time.sleep(seconds)


@pytest.fixture(scope="session")
def test_image():
    return TEST_IMAGE


@pytest.fixture(scope="session")
def test_video():
    return TEST_VIDEO


@pytest.fixture(scope="session")
def test_audio():
    return TEST_AUDIO


@pytest.fixture(scope="function")
def test_file():
    """Create a temporary test file"""
    # can't use tmp_path fixture because the tmpfs filesystem doesn't support xattrs
    with NamedTemporaryFile(dir=os.getcwd(), prefix="tmp_") as test_file:
        yield test_file


@pytest.fixture(scope="function")
def test_file2():
    """Create a temporary test file"""
    # can't use tmp_path fixture because the tmpfs filesystem doesn't support xattrs
    with NamedTemporaryFile(dir=os.getcwd(), prefix="tmp_") as test_file:
        yield test_file


@pytest.fixture(scope="function")
def test_dir():
    """Create a temporary directory"""
    # can't use tmp_path fixture because the tmpfs filesystem doesn't support xattrs
    with TemporaryDirectory(dir=os.getcwd(), prefix="tmp_") as test_dir:
        yield test_dir


def value_for_type(
    type_: type,
) -> t.Union[
    str, float, bool, datetime.datetime, t.List[str], t.List[datetime.datetime]
]:
    """Get test value for a given metadata attribute type"""
    if type_ == "str":
        return "Hello World"
    elif type_ == "float":
        return 42.0
    elif type_ == "bool":
        return True
    elif type_ == "datetime.datetime":
        return datetime.datetime(1995, 5, 31, 0, 0, 0)
    elif type_ == "list":
        return ["a", "b"]
    elif type_ == "list[datetime.datetime]":
        return [datetime.datetime(1995, 5, 31, 0, 0, 0)]
    else:
        raise ValueError(f"Unknown type: {type_}")



"""Test osxmetadata get/set com.apple.FinderInfo xattr methods"""

import pytest

from osxmetadata import (
    ALL_ATTRIBUTES,
    OSXMetaData,
    Tag,
    _kFinderColor,
    _kFinderInfo,
    _kFinderStationeryPad,
)
from osxmetadata.constants import (
    FINDER_COLOR_BLUE,
    FINDER_COLOR_GREEN,
    FINDER_COLOR_NONE,
)
from .conftest import snooze


def test_stationerypad(test_file):
    """test get/set stationerypad methods"""

    attribute = "com.apple.metadata:kMDItemComment"
    value = "This is my comment"

    md = OSXMetaData(test_file.name)
    md.stationerypad = True
    snooze()
    assert md.stationerypad == True

    md.set(_kFinderStationeryPad, False)
    snooze()
    assert md.get(_kFinderStationeryPad) == False


def test_finderinfo(test_file):
    """Test finderinfo attribute to get raw bytes"""

    md = OSXMetaData(test_file.name)
    snooze()
    assert len(md.finderinfo) == 32

    assert (
        md.finderinfo
        == b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    )
    md.stationerypad = True
    snooze()
    assert (
        md.finderinfo
        == b"\x00\x00\x00\x00\x00\x00\x00\x00\x08\x00\x00\x00\x00\x00\x00\x00"
        + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    )


def test_findercolor(test_file):
    """Test findercolor attribute to get/set color in the FinderInfo field"""

    md = OSXMetaData(test_file.name)
    assert md.findercolor == FINDER_COLOR_NONE
    md.findercolor = FINDER_COLOR_GREEN
    snooze()
    assert md.findercolor == FINDER_COLOR_GREEN

    # test that setting a tag color also sets the FinderInfo color
    md.findercolor = FINDER_COLOR_NONE
    md.tags = [Tag("Blue", color=FINDER_COLOR_BLUE)]
    snooze()
    assert md.findercolor == FINDER_COLOR_BLUE


def test_all_attributes():
    """Test that all Finder Info attributes are in ALL_ATTRIBUTES"""

    assert _kFinderStationeryPad in ALL_ATTRIBUTES
    assert _kFinderInfo in ALL_ATTRIBUTES
    assert _kFinderColor in ALL_ATTRIBUTES



"""Test exceptions raised by OSXMetaData"""

import datetime

import pytest

from osxmetadata import OSXMetaData


def test_get_invalid_attribute(test_file):
    """Test get invalid attribute"""
    md = OSXMetaData(test_file.name)
    with pytest.raises(AttributeError):
        md.invalid_attribute


def test_set_invalid_attribute(test_file):
    """Test set invalid attribute"""
    md = OSXMetaData(test_file.name)
    with pytest.raises(AttributeError):
        md.invalid_attribute = "value"


def test_set_readonly_attribute(test_file):
    """Test set readonly attribute"""
    md = OSXMetaData(test_file.name)
    with pytest.raises(AttributeError):
        md.kMDItemDateAdded = datetime.datetime.now()


def test_get_invalid_key(test_file):
    """Test get invalid key"""
    md = OSXMetaData(test_file.name)
    with pytest.raises(KeyError):
        md["invalid_key"]


def test_set_invalid_key(test_file):
    """Test set invalid key"""
    md = OSXMetaData(test_file.name)
    with pytest.raises(KeyError):
        md["invalid_key"] = "value"


def test_set_readonly_key(test_file):
    """Test set readonly key"""
    md = OSXMetaData(test_file.name)
    with pytest.raises(KeyError):
        md["kMDItemDateAdded"] = datetime.datetime.now()



""" Python package to read and write various Mac OS X metadata 
    such as tags/keywords and Finder comments from files """

from ._version import __version__
from .attribute_data import (
    MDIMPORTER_ATTRIBUTE_DATA,
    MDITEM_ATTRIBUTE_DATA,
    MDITEM_ATTRIBUTE_READ_ONLY,
    MDITEM_ATTRIBUTE_SHORT_NAMES,
    NSURL_RESOURCE_KEY_DATA,
)
from .constants import (
    FINDER_COLOR_BLUE,
    FINDER_COLOR_GRAY,
    FINDER_COLOR_GREEN,
    FINDER_COLOR_NONE,
    FINDER_COLOR_ORANGE,
    FINDER_COLOR_PURPLE,
    FINDER_COLOR_RED,
    FINDER_COLOR_YELLOW,
)
from .finder_info import _kFinderColor, _kFinderInfo, _kFinderStationeryPad
from .finder_tags import Tag, _kMDItemUserTags
from .mditem import MDItemValueType
from .osxmetadata import ALL_ATTRIBUTES, ASDICT_ATTRIBUTES, OSXMetaData

# add metadata attribute constants such as kMDItemFinderComment and NSURLTagNamesKey to module namespace
for constant in MDITEM_ATTRIBUTE_DATA.keys():
    globals()[constant] = constant
for constant in MDIMPORTER_ATTRIBUTE_DATA.keys():
    globals()[constant] = constant
for constant in NSURL_RESOURCE_KEY_DATA.keys():
    globals()[constant] = constant


__all__ = [
    "ALL_ATTRIBUTES",
    "ASDICT_ATTRIBUTES",
    "FINDER_COLOR_BLUE",
    "FINDER_COLOR_GRAY",
    "FINDER_COLOR_GREEN",
    "FINDER_COLOR_NONE",
    "FINDER_COLOR_ORANGE",
    "FINDER_COLOR_PURPLE",
    "FINDER_COLOR_RED",
    "FINDER_COLOR_YELLOW",
    "MDIMPORTER_ATTRIBUTE_DATA",
    "MDITEM_ATTRIBUTE_DATA",
    "MDITEM_ATTRIBUTE_READ_ONLY",
    "MDITEM_ATTRIBUTE_SHORT_NAMES",
    "MDItemValueType",
    "NSURL_RESOURCE_KEY_DATA",
    "OSXMetaData",
    "Tag",
    "__version__",
    "_kFinderColor",
    "_kFinderInfo",
    "_kFinderStationeryPad",
    "_kMDItemUserTags",
    *MDIMPORTER_ATTRIBUTE_DATA.keys(),
    *MDITEM_ATTRIBUTE_DATA.keys(),
    *NSURL_RESOURCE_KEY_DATA.keys(),
]



"""CLI for osxmetadata"""

import datetime
import glob
import json
import logging
import os
import os.path
import pathlib
import typing as t

import click

from osxmetadata import (
    ALL_ATTRIBUTES,
    MDIMPORTER_ATTRIBUTE_DATA,
    MDITEM_ATTRIBUTE_DATA,
    MDITEM_ATTRIBUTE_READ_ONLY,
    MDITEM_ATTRIBUTE_SHORT_NAMES,
    OSXMetaData,
    Tag,
    __version__,
    _kFinderColor,
    _kFinderInfo,
    _kFinderStationeryPad,
    _kMDItemUserTags,
)
from osxmetadata.backup import get_backup_dict, load_backup_file, write_backup_file
from osxmetadata.constants import _COLORNAMES_LOWER, _TAGS_NAMES, FINDER_COLOR_NONE
from osxmetadata.finder_info import str_to_finder_color
from osxmetadata.finder_tags import tag_factory
from osxmetadata.mditem import str_to_mditem_type

# TODO: how is metadata on symlink handled?
# should symlink be resolved before gathering metadata?
# currently, symlinks are resolved before handling metadata but not sure this is the right behavior
# in 10.13.6: synlinks appear to inherit tags but not Finder Comments:
#   e.g. changes to tags in a file changes tags in the symlink but symlink can have it's own finder comment
#   Finder aliases inherit neither
# TODO: add selective restore (e.g only restore files matching command line path)
#   e.g osxmetadata -r meta.json *.pdf

# TODO: fix output of str_to_mditem_type to be more helpful: ValueError: Invalid isoformat string: '2022-10-6'
# also wrap in try/except and print error message

_SHORT_NAME_WIDTH = (
    max(len(x["short_name"]) for x in MDITEM_ATTRIBUTE_DATA.values()) + 1
)
_LONG_NAME_WIDTH = max(len(x["name"]) for x in MDITEM_ATTRIBUTE_DATA.values()) + 1

BACKUP_FILENAME = ".osxmetadata.json"


def get_writeable_attributes() -> t.List[str]:
    """Return a list of writeable attributes"""
    no_write = ["kMDItemContentCreationDate", "kMDItemContentModificationDate"]
    write = [
        *MDITEM_ATTRIBUTE_DATA.keys(),
        _kFinderColor,
        _kFinderStationeryPad,
        _kMDItemUserTags,
    ]
    return [
        attr
        for attr in write
        if attr not in MDITEM_ATTRIBUTE_READ_ONLY and attr not in no_write
    ]


WRITABLE_ATTRIBUTES = get_writeable_attributes()


def get_attribute_type(attr: str) -> t.Optional[str]:
    """Get the type of an attribute

    Args:
        attr: attribute name

    Returns:
        type of attribute as string or None if type is not known
    """
    if attr in MDITEM_ATTRIBUTE_SHORT_NAMES:
        attr = MDITEM_ATTRIBUTE_SHORT_NAMES[attr]
    return (
        "list"
        if attr in _TAGS_NAMES
        else (
            MDITEM_ATTRIBUTE_DATA[attr]["python_type"]
            if attr in MDITEM_ATTRIBUTE_DATA
            else (
                "int"
                if attr == _kFinderColor
                else "bool" if attr == _kFinderStationeryPad else None
            )
        )
    )


def get_attribute_name(attr: str) -> str:
    """Get the long name of an attribute

    Args:
        attr: attribute name

    Returns:
        long name of attribute or attr if long name is not known
    """
    if attr in MDITEM_ATTRIBUTE_SHORT_NAMES:
        return MDITEM_ATTRIBUTE_SHORT_NAMES[attr]
    elif attr in MDITEM_ATTRIBUTE_DATA:
        return attr
    elif attr in MDIMPORTER_ATTRIBUTE_DATA:
        return attr
    elif attr in _TAGS_NAMES:
        return _kMDItemUserTags
    elif attr in [_kFinderColor, _kFinderStationeryPad]:
        return attr
    else:
        raise ValueError(f"Unknown attribute: {attr}")


def value_to_str(value) -> str:
    """Convert a metadata value to str suitable for printing to terminal"""
    if isinstance(value, str):
        return value
    elif value is None:
        return "(null)"
    if isinstance(value, datetime.datetime):
        return value.isoformat()
    elif isinstance(value, (list, tuple)):
        if not value:
            return "(empty list)"
        if isinstance(value[0], str):
            return ", ".join(value)
        elif isinstance(value[0], Tag):
            return ", ".join(f"{x.name}: {x.color}" for x in value)
        elif isinstance(value[0], datetime.datetime):
            return ", ".join(x.isoformat() for x in value)
        else:
            return ", ".join(str(x) for x in value)
    else:
        return str(value)


def str_to_bool(value: str) -> bool:
    """Convert str value to bool:

    Args:
        value: str value to convert to bool

    Returns: True if str value is "True" or "true", or non-zero int; False otherwise
    """
    try:
        return bool(int(value))
    except ValueError:
        return value.lower() == "true"


def get_attributes_to_wipe(md: OSXMetaData) -> t.List[str]:
    """Get list of non-null metadata attributes on a file that can be wiped"""

    attribute_list = []
    for attr in WRITABLE_ATTRIBUTES:
        if value := md.get(attr):
            attribute_list.append(attr)
    return attribute_list


def md_wipe_metadata(md: OSXMetaData, verbose: bool = False):
    """Wipe metadata attributes on a file

    Args:
        md: OSXMetaData object for file
        attributes: list of attributes to wipe
        verbose: if True, print verbose output
    """
    attr_list = get_attributes_to_wipe(md)
    filepath = md.path
    if verbose:
        if attr_list:
            click.echo(f"Wiping metadata from {filepath}")
        else:
            click.echo(f"No metadata to wipe from {filepath}")
    for attr in attr_list:
        try:
            if verbose:
                click.echo(f"  Wiping {attr} from {filepath}")
            md.set(attr, None)
        except AttributeError:
            if verbose:
                click.echo(
                    f"  Unknown attribute {attr} on {filepath}, skipping", err=True
                )


def validate_attribute_names(attributes: t.Union[t.Tuple[str], t.Tuple[str, str]]):
    """Validate attribute names as returned by click option parsing:

    Args:
        attributes: tuple of attribute names or tuple of attribute name and value

    Returns:
        True if valid, raises click.BadParameter if not
    """

    for attr in attributes:
        if isinstance(attr, tuple):
            attr = attr[0]
        if attr not in ALL_ATTRIBUTES:
            raise click.BadParameter(f"Invalid attribute name: {attr}")


def get_attribute_names(attribute: str) -> t.Tuple[str, str]:
    """Get the name and short name for a metadata attribute

    Args:
        attribute: attribute name or short name

    Returns:
        tuple of attribute name and short name
    """
    if attribute in MDITEM_ATTRIBUTE_DATA:
        attribute = MDITEM_ATTRIBUTE_DATA[attribute]
        short_name = attribute["short_name"]
        name = attribute["name"]
    elif attribute in MDITEM_ATTRIBUTE_SHORT_NAMES:
        short_name = attribute
        name = MDITEM_ATTRIBUTE_SHORT_NAMES[attribute]
    elif attribute in MDIMPORTER_ATTRIBUTE_DATA:
        attribute = MDIMPORTER_ATTRIBUTE_DATA[attribute]
        short_name = attribute["name"]
        name = attribute["name"]
    elif attribute in [_kFinderInfo, _kFinderColor, _kFinderStationeryPad]:
        short_name = attribute
        name = attribute
    elif attribute in _TAGS_NAMES:
        short_name = "tags"
        name = _kMDItemUserTags
    else:
        raise ValueError(f"Unknown attribute: {attribute}")

    return name, short_name


def md_copyfrom_metadata(md: OSXMetaData, copyfrom: str, verbose: bool = False):
    """Copy metadata attributes to a file from another file

    Args:
        md: OSXMetaData object for destination file
        copyfrom: path to source file
        verbose: if True, print verbose output
    """
    if verbose:
        click.echo(f"Copying attributes from {copyfrom}")
    src_md = OSXMetaData(copyfrom)
    for attr in WRITABLE_ATTRIBUTES:
        if value := src_md.get(attr):
            if verbose:
                click.echo(f"  Copying {attr}")
            md.set(attr, value)


def md_clear_metadata(
    md: OSXMetaData, tr, attributes: t.List[str], verbose: bool = False
):
    """Clear metadata attributes on a file

    Args:
        md: OSXMetaData object for file
        attributes: list of attributes to clear
        verbose: if True, print verbose output
    """
    for attr in attributes:
        if verbose:
            click.echo(f"Clearing {attr}")
        if not md.get(attr):
            if verbose:
                click.echo(f"  {attr} is already empty on {md.path}")
            continue
        md.set(attr, None)


def md_set_metadata_with_error(
    md: OSXMetaData, metadata: t.Tuple[t.Tuple[str, str]], verbose: bool
) -> t.Optional[str]:
    """Set metadata for OSXMetaData object, return error message if any

    Args:
        md: OSXMetaData object
        metadata: tuple of tuples of (attribute, value) as returned by click parser
        verbose: if True, print metadata being set

    Returns:
        None if successful, else error message
    """

    attr_dict = {}
    for item in metadata:
        attr, val = item
        val = val or None

        # Convert attribute shortcut name to long name if necessary
        attr = get_attribute_name(attr)

        if attr in _TAGS_NAMES:
            val = tag_factory(val) if val else None
        elif attr == _kFinderColor:
            val = str_to_finder_color(val)
        elif attr == _kFinderStationeryPad:
            val = str_to_bool(val)
        elif attr in MDITEM_ATTRIBUTE_DATA:
            val = str_to_mditem_type(attr, val)
        else:
            return f"Invalid attribute: {attr}"

        if attr in attr_dict:
            attr_dict[attr].append(val)
        else:
            attr_dict[attr] = [val]

    for attribute, value in attr_dict.items():
        # if we got a list of values and attribute takes a list, set the list
        # otherwise, set the last value in the list
        if verbose:
            click.echo(f"Setting {attribute}={value}")
        if get_attribute_type(attribute) in ["list", "list[datetime.datetime]"]:
            # filter out any None values ([None] should be [])
            value = [v for v in value if v is not None]
            md.set(attribute, value)
        else:
            md.set(attribute, value[-1])

    return None


def md_append_metadata_with_error(
    md: OSXMetaData, metadata: t.List[t.Tuple[str, str]], verbose: bool
) -> t.Optional[str]:
    """Append metadata attributes on a file

    Args:
        md: OSXMetaData object for file
        metadata: list of tuples of (attribute, value) as returned by click parser
        verbose: if True, print verbose output

    Returns:
        None if successful, else error message
    """
    for attr, val in metadata:
        if verbose:
            click.echo(f"Appending {attr}={val}")

        # Convert attribute shortcut name to long name if necessary
        attr = get_attribute_name(attr)

        if attr in _TAGS_NAMES:
            value = tag_factory(val)
        elif attr in MDITEM_ATTRIBUTE_DATA:
            value = str_to_mditem_type(attr, val)
        else:
            # other types like _kFinderColor cannot be appended to
            return f"Invalid attribute: {attr}"

        attr_type = get_attribute_type(attr)

        if attr_type in ["list", "list[datetime.datetime]"]:
            new_value = md.get(attr) or []
            if value not in new_value:
                new_value.append(value)
                md.set(attr, new_value)
            elif verbose:
                click.echo(f"  {attr} already contains {val}")
        elif attr_type == "str":
            new_value = md.get(attr) or ""
            md.set(attr, new_value + value)
        else:
            return f"Attribute {attr} does not support appending"

    return None


def md_remove_metadata_with_error(
    md: OSXMetaData, metadata: t.List[t.Tuple[str, str]], verbose: bool
) -> t.Optional[str]:
    """ "Remove metadata attributes on a file

    Args:
        md: OSXMetaData object for file
        metadata: list of tuples of (attribute, value) as returned by click parser
        verbose: if True, print verbose output

    Returns:
        None if successful, else error message
    """
    for attr, val in metadata:
        attr_type = get_attribute_type(attr)
        if attr_type not in ["list", "list[datetime.datetime]"]:
            return f"remove is not a valid operation for single-value attribute {attr}"

        if attr in _TAGS_NAMES:
            val = md_tag_value_from_file(md, val)
            val = tag_factory(val)
        elif attr in MDITEM_ATTRIBUTE_DATA or attr in MDITEM_ATTRIBUTE_SHORT_NAMES:
            val = str_to_mditem_type(attr, val)
        else:
            return f"Invalid attribute: {attr}"

        new_value = md.get(attr) or []
        new_value = [v for v in new_value if v != val]

        if verbose:
            click.echo(f"Removing {val} from {attr}")
        try:
            md.set(attr, new_value)
        except KeyError as e:
            raise e


def md_tag_value_from_file(md: OSXMetaData, value: str) -> str:
    """Given a tag value, return the tag + color if tag value contains color.
    If not, check if file has the same tag and if so, return the tag + color from the file

    Returns the new tag value
    """
    values = value.split(",")
    if len(values) > 2:
        raise ValueError(f"More than one value found after comma: {value}")
    if len(values) == 2:
        return value
    if file_tags := md.get(_kMDItemUserTags):
        for tag in file_tags:
            if tag.name.lower() == value.lower():
                return f"{value},{tag.color}"
    return value


def md_mirror_metadata_with_error(
    md: OSXMetaData, attributes: t.Tuple[t.Tuple[str, str]], verbose: bool
) -> t.Optional[str]:
    """Mirror metadata attributes on a file

    Args:
        md: OSXMetaData object for file
        attributes: tuple of attribute tuples to mirror as returned by click parser for --mirror
        verbose: if True, print verbose output

    Returns:
        None if successful, else error message
    """
    for item in attributes:
        attr1, attr2 = item
        if verbose:
            click.echo(f"Mirroring {attr1} {attr2}")

        attr_type1 = get_attribute_type(attr1)
        attr_type2 = get_attribute_type(attr2)

        if attr_type1 != attr_type2:
            return f"Attributes {attr1} and {attr2} are not compatible"

        if attr_type1 in ["list", "list[datetime.datetime]"]:
            value1 = md.get(attr1) or []
            value2 = md.get(attr2) or []
            if attr1 in _TAGS_NAMES and attr2 not in _TAGS_NAMES:
                # might be mirroring a keyword to a tag
                # convert non-tags to tags
                value2_tags = [tag_factory(v) for v in value2]
                value1 = value1 + value2_tags
                value2 = value2 + [v.name for v in value1 if v.name not in value2]
                md.set(attr1, value1)
                md.set(attr2, value2)
            elif attr2 in _TAGS_NAMES and attr1 not in _TAGS_NAMES:
                # might be mirroring a tag to a keyword
                # convert tags to non-tags
                value1_tags = [tag_factory(v) for v in value1]
                value2 = value2 + value1_tags
                value1 = value1 + [v.name for v in value2 if v.name not in value1]
                md.set(attr1, value1)
                md.set(attr2, value2)
            elif value1 != value2:
                new_value = value1 + [v for v in value2 if v not in value1]
                md.set(attr1, new_value)
                md.set(attr2, new_value)
        else:
            md.set(attr1, md.get(attr2))

        return None


def md_list_metadata_with_error(md: OSXMetaData, json_: bool) -> t.Optional[str]:
    """List metadata attributes on a file

    Args:
        md: OSXMetaData object for file
        json_: if True, print output as JSON
        verbose: if True, print verbose output

    Returns:
        None if successful, else error message
    """
    if json_:
        json_str = md.to_json()
        click.echo(json_str)
        return

    # print in readable format, not json
    click.echo(f"{md.path}:")
    for attr in sorted(md.asdict()):
        try:
            value = md.get(attr)
            if value is None or value == "" or value == []:
                continue
            value = value_to_str(value)
        except Exception as e:
            click.echo(
                f"{'Error loading attribute':{_SHORT_NAME_WIDTH}}{attr:{_LONG_NAME_WIDTH}}: {e}",
                err=True,
            )
        else:
            try:
                name, short_name = get_attribute_names(attr)
            except ValueError:
                click.echo(
                    f"{'UNKNOWN ATTRIBUTE':{_SHORT_NAME_WIDTH}}{attr:{_LONG_NAME_WIDTH}} = THIS ATTRIBUTE NOT HANDLED",
                    err=True,
                )
            else:
                click.echo(
                    f"{short_name:{_SHORT_NAME_WIDTH}}{name:{_LONG_NAME_WIDTH}} = {value}"
                )


def md_get_metadata_with_error(
    md: OSXMetaData, attributes: t.Tuple[str], json_: bool
) -> t.Optional[str]:
    """Get metadata attribute on a file

    Args:
        md: OSXMetaData object for file
        attributes: tuple of attributes to get as returned by click parser for --get
        json_: if True, print output as JSON

    Returns:
        None if successful, else error message
    """
    data = {}
    if json_:
        data["_version"] = __version__
        data["_filepath"] = md.path
        data["_filename"] = os.path.basename(md.path)
    for attr in attributes:
        try:
            value = md.get(attr)
            attr_type = get_attribute_type(attr)
            if json_ and attr_type in ["list", "list[datetime.datetime]"]:
                # preserve lists for json output and convert datetimes if needed
                if attr_type == "list[datetime.datetime]":
                    # value could be 'None' if attribute not set
                    value = [v.isoformat() for v in value] if value else []
            else:
                value = value_to_str(value)
        except Exception as e:
            return f"Error loading attribute {attr}: {e}"
        else:
            try:
                name, short_name = get_attribute_names(attr)
            except ValueError:
                return f"UNKNOWN ATTRIBUTE {attr}: THIS ATTRIBUTE NOT HANDLED"
            else:
                if json_:
                    data[name] = value
                else:
                    click.echo(
                        f"{short_name:{_SHORT_NAME_WIDTH}}{name:{_LONG_NAME_WIDTH}} = {value}"
                    )
    if json_:
        json_str = json.dumps(data)
        click.echo(json_str)


def validate_mirror_attributes_with_error(mirror: t.Tuple[t.Tuple[str, str]]):
    """Validate mirror attributes"""
    for item in mirror:
        attr1, attr2 = item

        validate_attribute_names((attr1, attr2))
        attr1 = get_attribute_name(attr1)
        attr2 = get_attribute_name(attr2)

        # avoid self mirroring
        if attr1 == attr2:
            return f"cannot mirror the same attribute: {attr1} {attr2}"

        # check type compatibility
        if get_attribute_type(attr1) != get_attribute_type(attr2):
            # can only mirror compatible attributes
            return f"Cannot mirror {attr1}, {attr2}: incompatible types"

    return None


def md_backup_metadata(filepath: str, backup_file: str, verbose: bool):
    """Backup metadata from file

    Args:
        filepath: path to file
        backup_file: path to backup file
        verbose: if True, print verbose output
    """
    # TODO: this is ripe for refactoring with a sqlite database
    # Currently, the code writes the entire backup database each time a file is processed
    if verbose:
        click.echo(f"  Backing up attribute data for {filepath}")
    # load the file if it exists, merge new data, then write out the file again
    backup_data = load_backup_file(backup_file) if backup_file.is_file() else {}
    backup_dict = get_backup_dict(filepath)
    backup_data[pathlib.Path(filepath).name] = backup_dict
    write_backup_file(backup_file, backup_data)


def md_restore_metadata(filepath: str, backup_file: str, verbose: bool):
    """Restore metadata from backup file

    Args:
        filepath: path to file to restore metadata for
        backup_file: path to backup file
        verbose: if True, print verbose output
    """

    try:
        backup_data = load_backup_file(backup_file)
        attr_dict = backup_data[pathlib.Path(filepath).name]
        if verbose:
            click.echo(f"  Restoring attribute data for {filepath}")
        md = OSXMetaData(filepath)
        for attr, value in attr_dict.items():
            if attr not in WRITABLE_ATTRIBUTES:
                continue
            if not value:
                # TODO: should values be set to None on restore if they were None in the backup?
                continue
            attr_type = get_attribute_type(attr)
            if attr in _TAGS_NAMES:
                value = [Tag(v[0], v[1]) for v in value]
            if attr_type == "datetime.datetime":
                value = datetime.datetime.fromisoformat(value)
            elif attr_type == "list[datetime.datetime]":
                value = [datetime.datetime.fromisoformat(v) for v in value]
            md.set(attr, value)
    except FileNotFoundError:
        click.echo(
            f"Missing backup file {backup_file} for {filepath}, skipping restore",
            err=True,
        )
    except KeyError:
        if verbose:
            click.echo(f"  Skipping restore for file {filepath}: not in backup file")


# Click CLI object & context settings
class CLI_Obj:
    def __init__(self, debug=False, files=None):
        self.debug = debug
        self.files = files


class MyClickCommand(click.Command):
    """Custom click.Command that overrides get_help() to show additional info"""

    def get_help(self, ctx):
        help_text = super().get_help(ctx)
        formatter = click.HelpFormatter()

        # build help text from all the attribute names
        # passed to click.HelpFormatter.write_dl for formatting
        attr_tuples = [("Short Name", "Description")]
        for attr in sorted(set(MDITEM_ATTRIBUTE_DATA.keys())):

            # get short and long name
            short_name = MDITEM_ATTRIBUTE_DATA[attr]["short_name"]
            long_name = MDITEM_ATTRIBUTE_DATA[attr]["name"]
            constant = MDITEM_ATTRIBUTE_DATA[attr]["xattr_constant"]

            # get help text
            description = MDITEM_ATTRIBUTE_DATA[attr]["description"]
            type_ = MDITEM_ATTRIBUTE_DATA[attr]["help_type"]
            attr_help = f"{long_name}; {constant}; {description}; {type_}"

            # add to list
            attr_tuples.append((short_name, attr_help))

        # add findercolor which isn't a standard kMDx item
        attr_tuples.append(
            (
                "findercolor",
                "findercolor; Finder color tag value. "
                + "The value can be either a number or the name of the color as follows: "
                + f"{', '.join([f'{colorid}: {color}' for color, colorid in _COLORNAMES_LOWER.items() if colorid != FINDER_COLOR_NONE])}; "
                + "integer or string.",
            )
        )
        attr_tuples = sorted(attr_tuples)

        formatter.write("\n\n")
        formatter.write_text(
            "Valid attributes for ATTRIBUTE: "
            + "Each attribute has a short name, a constant name, and a long constant name. "
            + "Any of these may be used for ATTRIBUTE"
        )
        formatter.write("\n")
        formatter.write_text('For example: --set findercomment "Hello world"')
        formatter.write_text('or:          --set kMDItemFinderComment "Hello world"')
        formatter.write_text(
            'or:          --set com.apple.metadata:kMDItemFinderComment "Hello world"'
        )
        formatter.write("\n")
        formatter.write_text(
            "Attributes that are strings can only take one value for --set; "
            + "--append will append to the existing value.  "
            + "Attributes that are arrays can be set multiple times to add to the array: "
            + "e.g. --set keywords 'foo' --set keywords 'bar' will set keywords to ['foo', 'bar']"
        )
        formatter.write("\n")
        formatter.write_text(
            "Options are executed in the following order regardless of order "
            + "passed on the command line: "
            + "restore, wipe, copyfrom, clear, set, append, remove, mirror, get, list, backup.  "
            + "--backup and --restore are mutually exclusive.  "
            + "Other options may be combined or chained together."
        )
        formatter.write("\n")
        formatter.write_text(
            "Finder tags (tags attribute) contain both a name and an optional color. "
            + "To specify the color, append comma + color name (e.g. 'red') after the "
            + "tag name.  For example --set tags Foo,red. "
            + "Valid color names are: "
            + f"{', '.join([color for color, colorid in _COLORNAMES_LOWER.items() if colorid != FINDER_COLOR_NONE])}. "
            + "If color is not specified but a tag of the same name has already been assigned a color "
            + "in the Finder, the same color will automatically be assigned. "
            + "See also findercolor."
        )
        formatter.write("\n")

        formatter.write_dl(attr_tuples)
        help_text += formatter.getvalue()
        return help_text


# All the command line options defined here
FILES_ARGUMENT = click.argument(
    "files", metavar="FILE", nargs=-1, type=click.Path(exists=True), required=True
)
WALK_OPTION = click.option(
    "--walk",
    "-w",
    is_flag=True,
    help="Walk directory tree, processing each file in the tree.",
    default=False,
)
JSON_OPTION = click.option(
    "--json",
    "-j",
    "json_",
    is_flag=True,
    help="Print output in JSON format, for use with --list and --get.",
    default=False,
)
DEBUG_OPTION = click.option(
    "--debug", required=False, is_flag=True, default=False, hidden=True
)
SET_OPTION = click.option(
    "--set",
    "-s",
    "set_",
    metavar="ATTRIBUTE VALUE",
    help="Set ATTRIBUTE to VALUE. "
    "If ATTRIBUTE is a multi-value attribute, such as keywords (kMDItemKeywords), "
    "you may specify --set multiple times to add to the array of values: "
    "'--set keywords foo --set keywords bar' will set keywords to ['foo', 'bar']. "
    "Not that this will overwrite any existing values for the attribute; see also --append.",
    nargs=2,
    multiple=True,
    required=False,
)
GET_OPTION = click.option(
    "--get",
    "-g",
    help="Get value of ATTRIBUTE.",
    metavar="ATTRIBUTE",
    nargs=1,
    multiple=True,
    required=False,
)
LIST_OPTION = click.option(
    "--list",
    "-l",
    "list_",
    help="List all metadata attributes for FILE.",
    is_flag=True,
    default=False,
)
CLEAR_OPTION = click.option(
    "--clear",
    "-c",
    help="Remove attribute from FILE.",
    metavar="ATTRIBUTE",
    nargs=1,
    multiple=True,
    required=False,
)
WIPE_OPTION = click.option(
    "--wipe",
    "-X",
    help="Wipe all metadata attributes from FILE.",
    is_flag=True,
    default=False,
    required=False,
)
APPEND_OPTION = click.option(
    "--append",
    "-a",
    metavar="ATTRIBUTE VALUE",
    help="Append VALUE to ATTRIBUTE; for multi-valued attributes, appends only if VALUE is not already present. "
    "May be used in combination with --set to add to an existing value: "
    "'--set keywords foo --append keywords bar' will set keywords to ['foo', 'bar'], "
    "overwriting any existing values for the attribute.",
    nargs=2,
    multiple=True,
    required=False,
)
REMOVE_OPTION = click.option(
    "--remove",
    "-r",
    metavar="ATTRIBUTE VALUE",
    help="Remove VALUE from ATTRIBUTE; only applies to multi-valued attributes.",
    nargs=2,
    multiple=True,
    required=False,
)
MIRROR_OPTION = click.option(
    "--mirror",
    "-m",
    metavar="ATTRIBUTE1 ATTRIBUTE2",
    help="Mirror values between ATTRIBUTE1 and ATTRIBUTE2 so that ATTRIBUTE1 = ATTRIBUTE2; "
    "for multi-valued attributes, merges values; for string attributes, sets ATTRIBUTE1 = ATTRIBUTE2 "
    "overwriting any value in ATTRIBUTE1.  "
    "For example: '--mirror keywords tags' sets tags and keywords to same values.",
    nargs=2,
    required=False,
    multiple=True,
)
BACKUP_OPTION = click.option(
    "--backup",
    "-B",
    help="Backup FILE attributes.  "
    "Backup file '.osxmetadata.json' will be created in same folder as FILE. "
    "Only backs up attributes known to osxmetadata unless used with --all.",
    is_flag=True,
    required=False,
    default=False,
)
RESTORE_OPTION = click.option(
    "--restore",
    "-R",
    help="Restore FILE attributes from backup file.  "
    "Restore will look for backup file '.osxmetadata.json' in same folder as FILE. "
    "Only restores attributes known to osxmetadata unless used with --all.",
    is_flag=True,
    required=False,
    default=False,
)
VERBOSE_OPTION = click.option(
    "--verbose",
    "-V",
    help="Print verbose output.",
    is_flag=True,
    default=False,
    required=False,
)
COPY_FROM_OPTION = click.option(
    "--copyfrom",
    "-f",
    metavar="SOURCE_FILE",
    help="Copy attributes from file SOURCE_FILE (only updates destination attributes that are not null in SOURCE_FILE).",
    type=click.Path(exists=True),
    nargs=1,
    multiple=False,
    required=False,
)
FILES_ONLY_OPTION = click.option(
    "--files-only",
    help="Do not apply metadata commands to directories themselves, only files in a directory.",
    is_flag=True,
    default=False,
    required=False,
)
PATTERN_OPTION = click.option(
    "--pattern",
    "-p",
    metavar="PATTERN",
    help="Only process files matching PATTERN; only applies to --walk. "
    "If specified, only files matching PATTERN will be processed as each directory is walked. "
    "May be used for than once to specify multiple patterns. "
    "For example, tag all *.pdf files in projectdir and subfolders with tag 'project': "
    "osxmetadata --append tags 'project' --walk projectdir/ --pattern '*.pdf'",
    multiple=True,
    required=False,
)


@click.command(cls=MyClickCommand)
@click.version_option(__version__, "--version", "-v")
@DEBUG_OPTION
@FILES_ARGUMENT
@WALK_OPTION
@JSON_OPTION
@WIPE_OPTION
@SET_OPTION
@LIST_OPTION
@CLEAR_OPTION
@APPEND_OPTION
@GET_OPTION
@REMOVE_OPTION
@MIRROR_OPTION
@BACKUP_OPTION
@RESTORE_OPTION
@VERBOSE_OPTION
@COPY_FROM_OPTION
@FILES_ONLY_OPTION
@PATTERN_OPTION
@click.pass_context
def cli(
    ctx,
    debug,
    files,
    walk,
    json_,
    wipe,
    set_,
    list_,
    clear,
    append,
    get,
    remove,
    mirror,
    backup,
    restore,
    verbose,
    copyfrom,
    files_only,
    pattern,
):
    """Read/write metadata from file(s)."""

    if debug:
        logging.disable(logging.NOTSET)

    # validate values for --set, --clear, --append, --get, --remove
    for attributes in [set_, append, remove, clear, get]:
        try:
            validate_attribute_names(attributes)
        except click.BadParameter as e:
            click.echo(e)
            ctx.exit(1)

    # check compatible types for mirror
    if mirror:
        if error := validate_mirror_attributes_with_error(mirror):
            click.echo(error, err=True)
            ctx.exit(1)

    # check that json_ only used with get or list_
    if json_ and not any([get, list_]):
        click.echo("--json can only be used with --get or --list", err=True)
        # click.echo("")  # add a new line before rest of help text
        # click.echo(ctx.get_help())
        ctx.exit(1)

    # can't backup and restore at once
    if backup and restore:
        click.echo("--backup and --restore cannot be used together", err=True)
        # click.echo("")  # add a new line before rest of help text
        # click.echo(ctx.get_help())
        ctx.exit(1)

    # loop through each file, process it, then do backup or restore if needed
    for filename in files:
        if not all([os.path.isdir(filename), walk, pattern]):
            process_files(
                ctx,
                [filename],
                json_,
                set_,
                append,
                remove,
                clear,
                get,
                list_,
                mirror,
                wipe,
                verbose,
                copyfrom,
                backup,
                restore,
                files_only,
            )

        if walk and os.path.isdir(filename):
            for root, dirnames, filenames in os.walk(filename):
                if pattern:
                    # only process files matching pattern
                    filepaths = []
                    for dirname in dirnames:
                        for matches in [
                            glob.glob(os.path.join(os.path.join(root, dirname), pat))
                            for pat in pattern
                        ]:
                            filepaths.extend(matches)
                else:
                    filepaths = [
                        os.path.join(root, fname) for fname in dirnames + filenames
                    ]
                process_files(
                    ctx,
                    filepaths,
                    json_,
                    set_,
                    append,
                    remove,
                    clear,
                    get,
                    list_,
                    mirror,
                    wipe,
                    verbose,
                    copyfrom,
                    backup,
                    restore,
                    files_only,
                )


def process_files(
    ctx,
    files,
    json_,
    set_,
    append,
    remove,
    clear,
    get,
    list_,
    mirror,
    wipe,
    verbose,
    copyfrom,
    backup,
    restore,
    files_only,
):
    """process list of files, calls process_single_file to process each file
    options processed in this order: wipe, copyfrom, clear, set, append, remove, mirror, get, list
    Note: expects all attributes passed in parameters to be validated as valid attributes
    """
    for filename in files:
        fpath = pathlib.Path(filename).resolve()
        backup_file = pathlib.Path(pathlib.Path(filename).parent) / BACKUP_FILENAME

        if files_only and fpath.is_dir():
            if verbose:
                click.echo(f"Skipping directory: {fpath}")
            continue

        if verbose:
            click.echo(f"Processing file: {fpath}")

        if restore:
            md_restore_metadata(fpath, backup_file, verbose)

        process_single_file(
            ctx,
            fpath,
            json_,
            set_,
            append,
            remove,
            clear,
            get,
            list_,
            mirror,
            wipe,
            verbose,
            copyfrom,
        )

        if backup:
            md_backup_metadata(fpath, backup_file, verbose)


def process_single_file(
    ctx,
    fpath,
    json_,
    set_,
    append,
    remove,
    clear,
    get,
    list_,
    mirror,
    wipe,
    verbose,
    copyfrom,
):
    """process a single file to apply the options
    options processed in this order: wipe, copyfrom, clear, set, append, remove, mirror, get, list
    Note: expects all attributes passed in parameters to be validated as valid attributes
    """

    md = OSXMetaData(fpath)

    if wipe:
        md_wipe_metadata(md, verbose)

    if copyfrom:
        # TODO: add option to clear existing attributes if copyfrom does not have them
        md_copyfrom_metadata(md, copyfrom, verbose)

    if clear:
        md_clear_metadata(md, fpath, clear, verbose)

    if set_:
        if error := md_set_metadata_with_error(md, set_, verbose):
            click.echo(error, err=True)
            ctx.exit(1)

    if append:
        if error := md_append_metadata_with_error(md, append, verbose):
            click.echo(error, err=True)
            ctx.exit(1)

    if remove:
        if error := md_remove_metadata_with_error(md, remove, verbose):
            click.echo(error, err=True)
            ctx.exit(1)

    if mirror:
        if error := md_mirror_metadata_with_error(md, mirror, verbose):
            click.echo(error, err=True)
            ctx.exit(1)

    if get:
        if error := md_get_metadata_with_error(md, get, json_):
            click.echo(error, err=True)
            ctx.exit(1)

    if list_:
        if error := md_list_metadata_with_error(md, json_):
            click.echo(error, err=True)
            ctx.exit(1)


if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter



"""Read _kMDItemUserTags (Finder tags) from extended attributes on a file.

Finder tags can be read/written with [NSURL getResourceValue:forKey:error:] and 
[NSURL setResourceValue:forKey:error:] using key NSURLTagNamesKey for tags and
NSURLLabelNumberKey for color (label) number. However, when a tag has a custom color,
only the tag name is returned by NSURLTagNamesKey and NSURLLabelNumberKey does not 
return the color of the custom tag. To get the color of a custom tag, you must read
the extended attribute _kMDItemUserTags. This module reads the extended attribute.
"""

import plistlib
import typing as t
from collections import namedtuple

import xattr
from Foundation import NSURL

from .constants import _COLORIDS, _COLORNAMES_LOWER, _MAX_FINDER_COLOR
from .nsurl_metadata import set_nsurl_metadata

NSURLTagNamesKey = "NSURLTagNamesKey"

Tag = namedtuple("Tag", ["name", "color"])

_kMDItemUserTags = "_kMDItemUserTags"
_kMDItemUserTagsXattr = "com.apple.metadata:_kMDItemUserTags"

__all__ = ["Tag", "_kMDItemUserTags", "get_finder_tags", "set_finder_tags"]


def split_tag_names_colors(tag_values: t.List[str]) -> t.List[Tag]:
    """Split tag name and color from tag value

    Args:
        tag_values: list of tag values

    Returns:
        list of Tag namedtuples
    """
    # tags are in format name\ncolor where color is an int 0-7
    # for example "Blue\n4"
    # I've seen custom tags that have more than one color in the color field though this shouldn't happen
    # so we'll just take the first color
    tags = []
    for tag_value in tag_values:
        tag_data = tag_value.split("\n")
        tag_name = tag_data[0]
        tag_color = int(tag_data[1]) if len(tag_data) > 1 else 0
        tags.append(Tag(tag_name, tag_color))
    return tags


def get_finder_tags(xattr_obj: xattr.xattr) -> t.List[Tag]:
    """Get Finder tags from extended attribute _kMDItemUserTags

    Args:
        xattr_obj: xattr.xattr object for file

    Returns:
        list of Finder tags
    """
    try:
        tag_data = xattr_obj[_kMDItemUserTagsXattr]
        # load the binary plist values
        tag_values = plistlib.loads(tag_data)
        tags = split_tag_names_colors(tag_values)
    except KeyError:
        tags = []
    return tags


def set_finder_tags(url: NSURL, tags: t.Optional[t.List[Tag]]):
    """Set Finder tags in extended attribute _kMDItemUserTags

    Args:
        url: NSURL for file
        tags: list of Finder tags
    """
    if tags is not None and not isinstance(tags, (list, tuple)):
        raise TypeError("tags must be a list or tuple of Tag namedtuples or None")
    if tags is not None and not all(isinstance(tag, Tag) for tag in tags):
        raise TypeError("tags items must be Tag namedtuples")

    # convert to list of strings in format name\ncolor
    tag_values = [f"{tag.name}\n{tag.color}" for tag in tags] if tags else []
    set_nsurl_metadata(url, NSURLTagNamesKey, tag_values)


def tag_factory(tag_str: str) -> Tag:
    """Creates a Tag namedtuple from a string in format 'name,color'

    Args:
        tag_str: (str) tag value in format: 'name,color' where name is the name of the tag and color specifies the color ID;

    Returns:
        Tag namedtuple

    Notes:
        The comma and color are optional; if not provided Tag will use color assigned in Finder or FINDER_COLOR_NONE if no color
            e.g. tag_factory("foo") -> Tag("foo", 0)
                 tag_factory("test,6") -> Tag("test", 6)
    """

    values = tag_str.split(",")
    if len(values) > 2:
        raise ValueError(f"More than one value found after comma: {tag_str}")
    name = values[0]
    name = name.lstrip().rstrip()

    if len(values) == 1:
        # got name only, check to see if name is also a color and if so assign it
        # TODO: this might not be right/desired in different locale settings
        # but I'm not sure yet how to get the non-English color names used by Finder
        try:
            colorid = _COLORNAMES_LOWER[name.lower()]
            return Tag(_COLORIDS[colorid], colorid)
        except KeyError:
            return Tag(name, 0)
    elif len(values) == 2:
        # got a color, is it a name or number
        color = values[1].lstrip().rstrip().lower()
        try:
            colorid = _COLORNAMES_LOWER[color]
        except KeyError as e:
            colorid = int(color)
            if colorid not in range(_MAX_FINDER_COLOR + 1):
                raise ValueError(
                    f"color must be in range 0 to {_MAX_FINDER_COLOR} inclusive: {colorid}"
                ) from e

        return Tag(name, colorid)



"""Task list for doit, used to build the project; run with `doit` or `doit list` to see commands"""

import osxmetadata


def task_update_readme():
    """Update README with CLI output"""
    return {"actions": ["cog -r README.md"]}


# def task_test():
#     """Run tests"""
#     return {"actions": ["pytest --doctest-glob=README.md tests/"]}


def task_docs():
    """Build docs"""
    return {"actions": ["mkdocs build"]}


def task_gh_docs():
    """Build docs and push to gh-pages"""
    return {
        "actions": [
            "mkdocs gh-deploy --force",
        ]
    }


def task_clean_build_files():
    """Clean out old build files"""
    return {
        "actions": ["rm -rf dist/", "rm -rf build/"],
    }


def task_build():
    """Build python project"""
    return {"actions": ["python3 -m build"]}


def task_build_exe():
    """Build exe with pyinstaller"""
    return {
        "actions": [
            'pyinstaller --onefile --hidden-import="pkg_resources.py2_warn" --name osxmetadata '
            "--add-data osxmetadata/attribute_data/audio_attributes.json:osxmetadata/attribute_data "
            "--add-data osxmetadata/attribute_data/common_attributes.json:osxmetadata/attribute_data "
            "--add-data osxmetadata/attribute_data/filesystem_attributes.json:osxmetadata/attribute_data "
            "--add-data osxmetadata/attribute_data/image_attributes.json:osxmetadata/attribute_data "
            "--add-data osxmetadata/attribute_data/mdimporter_constants.json:osxmetadata/attribute_data "
            "--add-data osxmetadata/attribute_data/nsurl_resource_keys.json:osxmetadata/attribute_data "
            "--add-data osxmetadata/attribute_data/video_attributes.json:osxmetadata/attribute_data "
            "cli.py"
        ]
    }


def task_zip_exe():
    """Zip executable for release"""
    version = osxmetadata._version.__version__
    return {
        "actions": [
            f"zip dist/osxmetadata_MacOS_exe_darwin_x64_v{version}.zip dist/osxmetadata",
            "rm dist/osxmetadata",
        ]
    }



"""Get/Set macOS metadata using MDItem* functions from CoreServices

Background: Apple provides MDItemCopyAttribute to get metadata from files:
https://developer.apple.com/documentation/coreservices/1427080-mditemcopyattribute?language=objc

but does not provide a documented way to set file metadata.

This package uses the undocumented functions MDItemSetAttribute, MDItemRemoveAttribute to do so.
"""

import datetime
import logging
import typing as t

import CoreFoundation
import CoreServices
import Foundation
import objc

from .attribute_data import (
    MDIMPORTER_ATTRIBUTE_DATA,
    MDITEM_ATTRIBUTE_DATA,
    MDITEM_ATTRIBUTE_SHORT_NAMES,
)

__all__ = [
    "get_mditem_metadata",
    "remove_mditem_metadata",
    "set_mditem_metadata",
    "set_or_remove_mditem_metadata",
    "str_to_mditem_type",
]

# Absolute time in macOS is measured in seconds relative to the absolute reference date of Jan 1 2001 00:00:00 GMT.
# Reference: https://developer.apple.com/documentation/corefoundation/1542812-cfdategetabsolutetime?language=objc
MACOS_TIME_DELTA = (
    datetime.datetime(2001, 1, 1, 0, 0) - datetime.datetime(1970, 1, 1, 0, 0)
).total_seconds()

MDItemValueType = t.Union[bool, str, float, t.List[str], datetime.datetime]

# load undocumented function MDItemSetAttribute
# signature: Boolean MDItemSetAttribute(MDItemRef, CFStringRef name, CFTypeRef attr);
# references:
# https://github.com/WebKit/WebKit/blob/5b8ad34f804c64c944ebe43c02aba88482c2afa8/Source/WTF/wtf/mac/FileSystemMac.MDItemSetAttribute
# https://pyobjc.readthedocs.io/en/latest/metadata/manual.html#objc.loadBundleFunctions
# signature of B@@@ translates to returns BOOL, takes 3 arguments, all objects
# In reality, the function takes references (pointers) to the objects, but pyobjc barfs if
# the function signature is specified using pointers.
# Specifying generic objects allows the bridge to convert the Python objects to the
# appropriate Objective C object pointers.


def MDItemSetAttribute(mditem, name, attr) -> bool:
    """dummy function definition"""
    ...


def MDItemRemoveAttribute(mditem, name):
    """ "dummy function definition"""
    ...


# This will load MDItemSetAttribute from the CoreServices framework into module globals
objc.loadBundleFunctions(
    CoreServices.__bundle__,
    globals(),
    [("MDItemSetAttribute", b"B@@@"), ("MDItemRemoveAttribute", b"B@@")],
)


def value_to_boolean(value: str) -> bool:
    """Convert string to boolean"""
    if value.lower() == "true":
        return True
    elif value.lower() == "false":
        return False
    elif value.isdigit():
        return bool(int(value))
    else:
        raise ValueError(f"Invalid boolean value: {value}")


def CFDate_to_datetime(cfdate: CoreFoundation.CFDateRef) -> datetime.datetime:
    """Convert CFDate to datetime.datetime"""
    return datetime.datetime.fromtimestamp(
        CoreFoundation.CFDateGetAbsoluteTime(cfdate) + MACOS_TIME_DELTA
    )


def NSDate_to_datetime(nsdate: Foundation.NSDate) -> datetime.datetime:
    """Convert NSDate to datetime.datetime"""
    return datetime.datetime.fromtimestamp(nsdate.timeIntervalSince1970())


def set_mditem_metadata(
    mditem: CoreServices.MDItemRef,
    attribute: str,
    value: MDItemValueType,
) -> bool:
    """Set file metadata using undocumented function MDItemSetAttribute

    mditem: MDItem object
    attribute: metadata attribute to set
    value: value to set attribute to; must match the type expected by the attribute (e.g. bool, str, List[str], float, datetime.datetime)

    Returns True if successful, False otherwise.
    """
    if isinstance(value, list):
        value = CoreFoundation.CFArrayCreate(
            None, value, len(value), CoreFoundation.kCFTypeArrayCallBacks
        )
    elif isinstance(value, datetime.datetime):
        value = CoreFoundation.CFDateCreate(None, value.timestamp() - MACOS_TIME_DELTA)
    return MDItemSetAttribute(
        mditem,
        attribute,
        value,
    )


def get_mditem_metadata(
    mditem: CoreServices.MDItemRef, attribute: str
) -> MDItemValueType:
    """Get file metadata using MDItemCopyAttribute

    mditem: MDItem object
    attribute: metadata attribute to get

    Returns value of attribute
    """
    value = CoreServices.MDItemCopyAttribute(mditem, attribute)
    if value is None:
        return None
    if attribute in MDITEM_ATTRIBUTE_DATA:
        attribute_data = MDITEM_ATTRIBUTE_DATA[attribute]
    elif attribute in MDIMPORTER_ATTRIBUTE_DATA:
        attribute_data = MDIMPORTER_ATTRIBUTE_DATA[attribute]
    else:
        raise ValueError(f"Unknown attribute: {attribute}")

    # Some attributes don't have a documented type
    attribute_type = attribute_data.get("python_type")
    try:
        if attribute_type == "bool":
            return bool(value)
        elif attribute_type == "str":
            return str(value)
        elif attribute_type == "float":
            return float(value)
        elif attribute_type == "list":
            # some attributes like kMDItemKeywords do not always follow the documented type
            # and can return a single comma-delimited string instead of a list (See #83)
            return (
                [x.strip() for x in str(value).split(",")]
                if isinstance(value, (objc.pyobjc_unicode, str))
                else [str(x) for x in value]
            )
        elif attribute_type == "datetime.datetime":
            return CFDate_to_datetime(value)
        elif attribute_type == "list[datetime.datetime]":
            return [CFDate_to_datetime(x) for x in value]
        # these are a hack but works for MDImporter attributes that don't have a documented type
        elif "__NSCFArray" in repr(type(value)):
            return [str(x) for x in value]
        elif "__NSArrayI" in repr(type(value)):
            return [str(x) for x in value]
        elif "__NSArrayM" in repr(type(value)):
            return [str(x) for x in value]
        elif "__NSTaggedDate" in repr(type(value)):
            return NSDate_to_datetime(value)
        elif isinstance(value, objc.pyobjc_unicode):
            return str(value)
        else:
            return value
    except ValueError:
        logging.warning(
            f"Failed to convert value ({value}) for attribute {attribute} to type {attribute_type}"
        )
        return None


def remove_mditem_metadata(mditem: CoreServices.MDItemRef, attribute: str):
    """Remove file metadata using undocumented function MDItemRemoveAttribute

    mditem: MDItem object
    attribute: metadata attribute to remove
    """
    MDItemRemoveAttribute(mditem, attribute)


def set_or_remove_mditem_metadata(
    mditem: CoreServices.MDItemRef,
    attribute: str,
    value: MDItemValueType,
) -> bool:
    """Set file metadata or remove metadata if value is None

    file: path to file
    attribute: metadata attribute to set
    value: value to set attribute to; must match the type expected by the attribute (e.g. bool, str, List[str], float, datetime.datetime)

    Returns True if successful, False otherwise.
    """
    if value is None:
        remove_mditem_metadata(mditem, attribute)
    else:
        set_mditem_metadata(mditem, attribute, value)


def str_to_mditem_type(attribute: str, value: str) -> MDItemValueType:
    """Convert string to type expected by MDItem attribute;

    Args:
        attribute: name of attribute
        value: string value to convert

    Returns: value converted to type expected by set_mditem_metadata

    Raises:
        ValueError: if value cannot be converted to expected type

    Notes:
        For example, kMDItemDueDate expects a datetime.datetime so this function will convert
        a string to a datetime.datetime using datetime.datetime.fromisoformat

        Caller is responsible for knowing if attribute is a list-type attribute; this method returns
        a single value, not a list for all attributes.
    """
    if attribute in MDITEM_ATTRIBUTE_DATA:
        attribute_data = MDITEM_ATTRIBUTE_DATA[attribute]
    elif attribute in MDITEM_ATTRIBUTE_SHORT_NAMES:
        attribute_data = MDITEM_ATTRIBUTE_DATA[MDITEM_ATTRIBUTE_SHORT_NAMES[attribute]]
    else:
        raise ValueError(f"Unknown attribute: {value}")

    attribute_type = attribute_data.get("python_type")
    if attribute_type == "str":
        return value
    if attribute_type == "bool":
        return bool(value)
    elif attribute_type == "float":
        return float(value)
    elif attribute_type == "list":
        # caller is responsible for knowing if attribute is a list-type attribute
        return value
    elif attribute_type == "datetime.datetime":
        return datetime.datetime.fromisoformat(value)
    elif attribute_type == "list[datetime.datetime]":
        # caller is responsible for knowing if attribute is a list-type attribute
        return datetime.datetime.fromisoformat(value)
    else:
        return value



"""Load metadata about metadata attributes from JSON files"""

from .load_attribute_data import (
    MDIMPORTER_ATTRIBUTE_DATA,
    MDITEM_ATTRIBUTE_AUDIO,
    MDITEM_ATTRIBUTE_COMMON,
    MDITEM_ATTRIBUTE_DATA,
    MDITEM_ATTRIBUTE_FILESYSTEM,
    MDITEM_ATTRIBUTE_IMAGE,
    MDITEM_ATTRIBUTE_READ_ONLY,
    MDITEM_ATTRIBUTE_SHORT_NAMES,
    MDITEM_ATTRIBUTE_VIDEO,
    NSURL_RESOURCE_KEY_DATA,
)

__all__ = [
    "MDIMPORTER_ATTRIBUTE_DATA",
    "MDITEM_ATTRIBUTE_AUDIO",
    "MDITEM_ATTRIBUTE_COMMON",
    "MDITEM_ATTRIBUTE_DATA",
    "MDITEM_ATTRIBUTE_FILESYSTEM",
    "MDITEM_ATTRIBUTE_IMAGE",
    "MDITEM_ATTRIBUTE_READ_ONLY",
    "MDITEM_ATTRIBUTE_SHORT_NAMES",
    "MDITEM_ATTRIBUTE_VIDEO",
    "NSURL_RESOURCE_KEY_DATA",
]



""" stand alone command line script for use with pyinstaller
    
    Note: This is *not* the cli that "python3 -m pip install osxmetadata" or "python setup.py install" would install;
    it's merely a wrapper around __main__.py to allow pyinstaller to work

    This script is built with `doit build_exe` and the resulting executable is zipped with `doit zip_exe` 
"""

from osxmetadata.__main__ import cli

if __name__ == "__main__":
    cli()



"""Test datetime handling """

import datetime
import os
import plistlib
import time

import pytest

from osxmetadata import OSXMetaData
from osxmetadata.datetime_utils import *


def test_datetime_attribute_naive(test_file):
    """Test datetime attribute with naive value"""

    os.environ["TZ"] = "US/Pacific"
    time.tzset()

    # naive datetime in local timezone
    duedate = datetime.datetime(2022, 10, 1, 1, 2, 3)
    md = OSXMetaData(test_file.name)
    md.kMDItemDueDate = duedate
    assert md.kMDItemDueDate == duedate

    # Apple stores dates as UTC, so we need to convert to UTC
    xattr_datetime = md.get_xattr(
        "com.apple.metadata:kMDItemDueDate", decode=plistlib.loads
    )
    duedate_utc = datetime_remove_tz(
        datetime_naive_to_local(duedate).astimezone(datetime.timezone.utc)
    )
    assert xattr_datetime == duedate_utc


def test_datetime_attribute_tz_aware(test_file):
    """Test datetime attribute with timezone aware value"""

    os.environ["TZ"] = "US/Pacific"
    time.tzset()

    # naive datetime in local timezone
    duedate = datetime_naive_to_local(datetime.datetime(2022, 10, 1, 1, 2, 3))
    md = OSXMetaData(test_file.name)
    md.kMDItemDueDate = duedate
    assert md.kMDItemDueDate == datetime_remove_tz(duedate)

    # Apple stores dates as UTC, so we need to convert to UTC
    xattr_datetime = md.get_xattr(
        "com.apple.metadata:kMDItemDueDate", decode=plistlib.loads
    )
    duedate_utc = datetime_remove_tz(duedate.astimezone(datetime.timezone.utc))
    assert xattr_datetime == duedate_utc



"""Test datetime_utils """

import datetime
import os
from datetime import date, timezone

import pytest

import osxmetadata.datetime_utils


def test_get_local_tz():
    os.environ["TZ"] = "US/Pacific"

    dt = datetime.datetime(2020, 9, 1, 21, 10, 00)
    tz = osxmetadata.datetime_utils.get_local_tz(dt)
    assert tz == datetime.timezone(offset=datetime.timedelta(seconds=-25200))

    dt = datetime.datetime(2020, 12, 1, 21, 10, 00)
    tz = osxmetadata.datetime_utils.get_local_tz(dt)
    assert tz == datetime.timezone(offset=datetime.timedelta(seconds=-28800))


def test_datetime_has_tz():
    tz = datetime.timezone(offset=datetime.timedelta(seconds=-28800))
    dt = datetime.datetime(2020, 9, 1, 21, 10, 00, tzinfo=tz)
    assert osxmetadata.datetime_utils.datetime_has_tz(dt)

    dt = datetime.datetime(2020, 9, 1, 21, 10, 00)
    assert not osxmetadata.datetime_utils.datetime_has_tz(dt)


def test_datetime_tz_to_utc():
    tz = datetime.timezone(offset=datetime.timedelta(seconds=-25200))
    dt = datetime.datetime(2020, 9, 1, 22, 6, 0, tzinfo=tz)
    utc = osxmetadata.datetime_utils.datetime_tz_to_utc(dt)
    assert utc == datetime.datetime(2020, 9, 2, 5, 6, 0, tzinfo=datetime.timezone.utc)


def test_datetime_remove_tz():
    os.environ["TZ"] = "US/Pacific"

    tz = datetime.timezone(offset=datetime.timedelta(seconds=-25200))
    dt = datetime.datetime(2020, 9, 1, 22, 6, 0, tzinfo=tz)
    dt = osxmetadata.datetime_utils.datetime_remove_tz(dt)
    assert dt == datetime.datetime(2020, 9, 1, 22, 6, 0)
    assert not osxmetadata.datetime_utils.datetime_has_tz(dt)


def test_datetime_naive_to_utc():
    dt = datetime.datetime(2020, 9, 1, 12, 0, 0)
    utc = osxmetadata.datetime_utils.datetime_naive_to_utc(dt)
    assert utc == datetime.datetime(2020, 9, 1, 12, 0, 0, tzinfo=datetime.timezone.utc)


def test_datetime_naive_to_local():
    os.environ["TZ"] = "US/Pacific"

    tz = datetime.timezone(offset=datetime.timedelta(seconds=-25200))
    dt = datetime.datetime(2020, 9, 1, 12, 0, 0)
    utc = osxmetadata.datetime_utils.datetime_naive_to_local(dt)
    assert utc == datetime.datetime(2020, 9, 1, 12, 0, 0, tzinfo=tz)


def test_datetime_utc_to_local():
    os.environ["TZ"] = "US/Pacific"

    tz = datetime.timezone(offset=datetime.timedelta(seconds=-25200))
    utc = datetime.datetime(2020, 9, 1, 19, 0, 0, tzinfo=datetime.timezone.utc)
    dt = osxmetadata.datetime_utils.datetime_utc_to_local(utc)
    assert dt == datetime.datetime(2020, 9, 1, 12, 0, 0, tzinfo=tz)


def test_datetime_utc_to_local_2():
    os.environ["TZ"] = "CEST"

    tz = datetime.timezone(offset=datetime.timedelta(seconds=7200))
    utc = datetime.datetime(2020, 9, 1, 19, 0, 0, tzinfo=datetime.timezone.utc)
    dt = osxmetadata.datetime_utils.datetime_utc_to_local(utc)
    assert dt == datetime.datetime(2020, 9, 1, 21, 0, 0, tzinfo=tz)


def test_datetime_to_new_tz():
    """Test datetime_to_new_tz"""
    tz = datetime.timezone(offset=datetime.timedelta(seconds=-25200))
    dt = datetime.datetime(2021, 10, 1, 0, 30, 0, tzinfo=tz)
    dt_new = osxmetadata.datetime_utils.datetime_to_new_tz(dt, 0)
    assert dt_new == datetime.datetime(
        2021, 10, 1, 7, 30, 0, tzinfo=datetime.timezone.utc
    )

    dt_new = osxmetadata.datetime_utils.datetime_to_new_tz(dt, 3600)
    tz_new = datetime.timezone(offset=datetime.timedelta(seconds=3600))
    assert dt_new == datetime.datetime(2021, 10, 1, 8, 30, 0, tzinfo=tz_new)


def test_utc_offset_seconds():

    dt_utc = datetime.datetime(2021, 9, 1, 0, 0, 0, 0, tzinfo=datetime.timezone.utc)
    assert osxmetadata.datetime_utils.utc_offset_seconds(dt_utc) == 0

    dt_pdt = datetime.datetime(
        2021, 9, 1, 0, 0, 0, 0, tzinfo=datetime.timezone(datetime.timedelta(hours=-7))
    )
    assert osxmetadata.datetime_utils.utc_offset_seconds(dt_pdt) == -25200



""" Test osxmetadata command line interface """

import datetime
import glob
import json
import os
import pathlib

import pytest
from click.testing import CliRunner

from osxmetadata import *
from osxmetadata import __version__
from osxmetadata.__main__ import BACKUP_FILENAME, cli
from osxmetadata.backup import load_backup_file

from .conftest import FINDER_COMMENT_SNOOZE, LONG_SNOOZE, snooze


def parse_cli_output(output):
    """Helper for testing

    Parse the CLI --list output and return value of all set attributes as dict
    """
    import re

    matches = re.findall(r"^(\w+)\s+.*\=\s+(.*)$", output, re.MULTILINE)
    return {match[0]: match[1] for match in matches}


def test_cli_list(test_file):
    """Test --list"""

    md = OSXMetaData(test_file.name)
    md.authors = ["John Doe"]
    md.findercomment = "Hello World"
    md.tags = [Tag("test", 0)]
    md.description = "This is a test file"

    snooze(FINDER_COMMENT_SNOOZE)

    runner = CliRunner()
    result = runner.invoke(
        cli,
        ["--list", test_file.name],
    )
    assert result.exit_code == 0
    output = parse_cli_output(result.stdout)
    assert output["description"] == "This is a test file"
    assert output["findercomment"] == "Hello World"
    assert output["authors"] == "John Doe"
    assert output["tags"] == "test: 0"


def test_cli_version():
    """Test --version"""

    runner = CliRunner()
    result = runner.invoke(cli, ["--version"])
    assert result.exit_code == 0
    assert __version__ in result.stdout


def test_cli_wipe(test_file):
    """Test --wipe"""

    md = OSXMetaData(test_file.name)
    md.authors = ["John Doe"]
    md.description = "This is a test file"

    snooze()
    runner = CliRunner()
    result = runner.invoke(
        cli,
        ["--wipe", test_file.name],
    )
    snooze()
    assert result.exit_code == 0
    assert not md.authors
    assert not md.description


def test_cli_set(test_file):
    """Test --set"""

    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "--set",
            "authors",
            "John Doe",
            "--set",
            "tags",
            "test,0",
            "--set",
            "description",
            "Hello World",
            "--set",
            "description",
            "Goodbye World",  # this should overwrite the previous value
            test_file.name,
        ],
    )
    snooze()
    assert result.exit_code == 0
    md = OSXMetaData(test_file.name)
    assert md.authors == ["John Doe"]
    assert md.tags == [Tag("test", 0)]
    assert md.description == "Goodbye World"


def test_cli_set_multi_keywords_1(test_file):
    """Test --set with multiple keywords (#83)"""

    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "--set",
            "keywords",
            "Foo",
            "--set",
            "keywords",
            "Bar",
            test_file.name,
        ],
    )
    snooze()
    assert result.exit_code == 0
    md = OSXMetaData(test_file.name)
    assert sorted(md.keywords) == ["Bar", "Foo"]


def test_cli_set_multi_keywords_2(test_file):
    """Test --set, --append with multiple keywords (#83)"""

    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "--set",
            "keywords",
            "Foo",
            "--append",
            "keywords",
            "Bar",
            test_file.name,
        ],
    )
    snooze()
    assert result.exit_code == 0
    md = OSXMetaData(test_file.name)
    assert sorted(md.keywords) == ["Bar", "Foo"]


def test_cli_clear(test_file):
    """Test --clear"""

    md = OSXMetaData(test_file.name)
    md.authors = ["John Doe"]
    md.description = "This is a test file"

    runner = CliRunner()
    result = runner.invoke(
        cli,
        ["--clear", "authors", test_file.name],
    )
    snooze()
    assert result.exit_code == 0
    assert not md.authors
    assert md.description == "This is a test file"


def test_cli_append(test_file):
    """Test --append"""

    md = OSXMetaData(test_file.name)
    md.authors = ["John Doe"]

    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "--append",
            "authors",
            "Jane Doe",
            "--append",  # append to empty attribute
            "tags",
            "test,0",
            test_file.name,
        ],
    )
    assert result.exit_code == 0
    assert md.authors == ["John Doe", "Jane Doe"]
    assert md.tags == [Tag("test", 0)]


def test_cli_set_then_append(test_file):
    """Test --set then --append"""

    md = OSXMetaData(test_file.name)
    md.authors = ["John Doe"]

    # set initial value
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "--set",
            "keywords",
            "foo",
            test_file.name,
        ],
    )
    assert result.exit_code == 0

    # set again and verify that it overwrites
    result = runner.invoke(
        cli,
        [
            "--set",
            "keywords",
            "bar",
            test_file.name,
        ],
    )
    assert result.exit_code == 0
    assert md.keywords == ["bar"]

    # append and verify that it appends
    result = runner.invoke(
        cli,
        [
            "--append",
            "keywords",
            "baz",
            test_file.name,
        ],
    )
    assert result.exit_code == 0
    assert sorted(md.keywords) == ["bar", "baz"]


def test_cli_get(test_file):
    """Test --get"""

    md = OSXMetaData(test_file.name)
    md.authors = ["John Doe"]
    md.description = "This is a test file"

    runner = CliRunner()
    result = runner.invoke(
        cli,
        ["--get", "authors", test_file.name],
    )
    assert result.exit_code == 0
    output = parse_cli_output(result.stdout)
    assert output["authors"] == "John Doe"


def test_cli_remove(test_file):
    """Test --remove"""

    md = OSXMetaData(test_file.name)
    md.authors = ["John Doe", "Jane Doe"]
    md.tags = [Tag("test", 0)]
    snooze()

    runner = CliRunner()
    result = runner.invoke(cli, ["--list", "--json", test_file.name])
    data = json.loads(result.output)
    assert sorted(data["kMDItemAuthors"]) == ["Jane Doe", "John Doe"]

    result = runner.invoke(
        cli,
        [
            "--remove",
            "authors",
            "John Doe",
            "--remove",
            "tags",
            "test,0",
            "--verbose",
            test_file.name,
        ],
    )
    assert result.exit_code == 0
    assert "Removing John Doe from authors" in result.output
    # for some reason this test fails without an additional delay
    # for the removed metadata to be updated on disk
    # without the additional delay, reading the metadata reads the previous value
    snooze(LONG_SNOOZE)

    result = runner.invoke(cli, ["--list", "--json", test_file.name])
    data = json.loads(result.output)
    assert data["kMDItemAuthors"] == ["Jane Doe"]


def test_cli_remove_tags_without_color(test_file):
    """Test --remove tags without specifying color (#106)"""

    runner = CliRunner()
    result = runner.invoke(cli, ["--set", "tags", ".Test,red", test_file.name])
    snooze(LONG_SNOOZE)

    md = OSXMetaData(test_file.name)
    assert md.tags == [Tag(".Test", 6)]

    result = runner.invoke(
        cli,
        ["--remove", "tags", ".Test", test_file.name],
    )
    assert result.exit_code == 0

    snooze(LONG_SNOOZE)
    md = OSXMetaData(test_file.name)
    assert not md.tags


def test_cli_mirror(test_file):
    """Test --mirror"""

    md = OSXMetaData(test_file.name)
    md.description = "This is a test file"

    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "--mirror",
            "comment",
            "description",
            test_file.name,
            "--verbose",
        ],
    )
    snooze()
    assert result.exit_code == 0
    assert "Mirroring" in result.stdout

    md = OSXMetaData(test_file.name)
    assert md.description == "This is a test file"


def test_cli_copyfrom(test_file, test_file2):
    """Test --copyfrom"""

    md = OSXMetaData(test_file.name)
    md.description = "This is a test file"

    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "--copyfrom",
            test_file.name,
            test_file2.name,
        ],
    )
    snooze()
    assert result.exit_code == 0

    md = OSXMetaData(test_file2.name)
    assert md.description == "This is a test file"


def test_cli_walk(test_dir):
    """test --walk"""

    dirname = pathlib.Path(test_dir)
    os.makedirs(dirname / "temp" / "subfolder1")
    os.makedirs(dirname / "temp" / "subfolder2")
    (dirname / "temp" / "temp1.txt").touch()
    (dirname / "temp" / "subfolder1" / "sub1.txt").touch()

    runner = CliRunner()
    result = runner.invoke(cli, ["--set", "tags", "test", "--walk", test_dir])
    snooze()
    assert result.exit_code == 0

    md = OSXMetaData(dirname / "temp" / "subfolder1" / "sub1.txt")
    assert md.tags == [Tag("test", 0)]

    md = OSXMetaData(dirname / "temp" / "subfolder2")
    assert md.tags == [Tag("test", 0)]


def test_cli_walk_files_only(test_dir):
    """test --walk with --files-only"""

    dirname = pathlib.Path(test_dir)
    os.makedirs(dirname / "temp" / "subfolder1")
    os.makedirs(dirname / "temp" / "subfolder2")
    (dirname / "temp" / "temp1.txt").touch()
    (dirname / "temp" / "subfolder1" / "sub1.txt").touch()

    runner = CliRunner()
    result = runner.invoke(
        cli, ["--set", "tags", "test", "--walk", "--files-only", test_dir]
    )
    snooze()
    assert result.exit_code == 0

    md = OSXMetaData(dirname / "temp" / "subfolder1" / "sub1.txt")
    assert md.tags == [Tag("test", 0)]

    md = OSXMetaData(dirname / "temp" / "subfolder2")
    assert not md.tags


def test_cli_walk_pattern(test_dir):
    """test --walk with --pattern"""

    dirname = pathlib.Path(test_dir)
    os.makedirs(dirname / "temp" / "subfolder1")
    os.makedirs(dirname / "temp" / "subfolder2")
    (dirname / "temp" / "temp1.txt").touch()
    (dirname / "temp" / "subfolder1" / "sub1.txt").touch()
    (dirname / "temp" / "subfolder1" / "sub1.pdf").touch()

    runner = CliRunner()
    result = runner.invoke(
        cli, ["--set", "tags", "test", "--walk", "--pattern", "*.pdf", test_dir]
    )
    assert result.exit_code == 0

    md = OSXMetaData(dirname / "temp" / "subfolder1" / "sub1.pdf")
    assert md.tags == [Tag("test", 0)]

    md = OSXMetaData(dirname / "temp" / "subfolder1" / "sub1.txt")
    assert not md.tags

    md = OSXMetaData(dirname / "temp" / "subfolder2")
    assert not md.tags


def test_cli_walk_pattern_2(test_dir):
    """test --walk with more than one --pattern"""

    dirname = pathlib.Path(test_dir)
    os.makedirs(dirname / "temp" / "subfolder1")
    os.makedirs(dirname / "temp" / "subfolder2")
    (dirname / "temp" / "temp1.txt").touch()
    (dirname / "temp" / "subfolder1" / "sub1.txt").touch()
    (dirname / "temp" / "subfolder1" / "sub1.pdf").touch()
    (dirname / "temp" / "subfolder2" / "sub2.jpg").touch()

    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "--set",
            "tags",
            "test",
            "--walk",
            "--pattern",
            "*.pdf",
            "--pattern",
            "*.jpg",
            test_dir,
        ],
    )
    assert result.exit_code == 0

    md = OSXMetaData(dirname / "temp" / "subfolder1" / "sub1.pdf")
    assert md.tags == [Tag("test", 0)]

    md = OSXMetaData(dirname / "temp" / "subfolder2" / "sub2.jpg")
    assert md.tags == [Tag("test", 0)]

    md = OSXMetaData(dirname / "temp" / "subfolder1" / "sub1.txt")
    assert not md.tags

    md = OSXMetaData(dirname / "temp" / "subfolder2")
    assert not md.tags


def test_cli_files_only(test_dir):
    """test --files-only without --walk"""

    dirname = pathlib.Path(test_dir)
    os.makedirs(dirname / "temp" / "subfolder1")
    os.makedirs(dirname / "temp" / "subfolder2")
    (dirname / "temp" / "temp1.txt").touch()
    (dirname / "temp" / "subfolder1" / "sub1.txt").touch()

    files = glob.glob(str(dirname / "temp" / "*"))

    runner = CliRunner()
    result = runner.invoke(cli, ["--set", "tags", "test", "--files-only", *files])
    assert result.exit_code == 0

    md = OSXMetaData(dirname / "temp" / "temp1.txt")
    assert md.tags == [Tag("test", 0)]

    md = OSXMetaData(dirname / "temp" / "subfolder1")
    assert not md.tags

    md = OSXMetaData(dirname / "temp" / "subfolder1" / "sub1.txt")
    assert not md.tags


def test_cli_backup_restore(test_dir):
    """Test --backup and --restore"""

    dirname = pathlib.Path(test_dir)
    test_file = dirname / "test_file.txt"
    test_file.touch()

    md = OSXMetaData(test_file)
    md.tags = [Tag("test", 0)]
    md.authors = ["John Doe", "Jane Doe"]
    md.wherefroms = ["http://www.apple.com"]
    md.downloadeddate = [datetime.datetime(2019, 1, 1, 0, 0, 0)]
    md.stationerypad = True

    runner = CliRunner()
    result = runner.invoke(cli, ["--backup", test_file.as_posix()])
    assert result.exit_code == 0

    # test the backup file was written and is readable
    backup_file = dirname / BACKUP_FILENAME
    assert backup_file.is_file()
    backup_data = load_backup_file(backup_file)
    assert backup_data[test_file.name]["stationerypad"] == True

    # wipe the data
    result = runner.invoke(cli, ["--wipe", test_file.as_posix()])
    snooze(LONG_SNOOZE)
    md = OSXMetaData(test_file)
    assert not md.tags
    assert not md.authors
    assert not md.stationerypad

    # restore the data
    result = runner.invoke(cli, ["--restore", test_file.as_posix()])
    snooze(LONG_SNOOZE)
    assert result.exit_code == 0
    assert md.tags == [Tag("test", 0)]
    assert md.authors == ["John Doe", "Jane Doe"]
    assert md.wherefroms == ["http://www.apple.com"]
    assert md.downloadeddate == [datetime.datetime(2019, 1, 1, 0, 0, 0)]
    assert md.stationerypad


def test_cli_backup_walk_pattern(test_dir):
    """test --backup --walk with --pattern"""

    dirname = pathlib.Path(test_dir)
    os.makedirs(dirname / "temp" / "subfolder1")
    os.makedirs(dirname / "temp" / "subfolder2")
    (dirname / "temp" / "temp1.txt").touch()
    (dirname / "temp" / "subfolder1" / "sub1.txt").touch()
    (dirname / "temp" / "subfolder1" / "sub1.pdf").touch()

    runner = CliRunner()
    result = runner.invoke(
        cli,
        ["--set", "tags", "test", "--walk", "--pattern", "*.pdf", "--backup", test_dir],
    )
    assert result.exit_code == 0

    md = OSXMetaData(dirname / "temp" / "subfolder1" / "sub1.pdf")
    assert md.tags == [Tag("test", 0)]

    backup_file = dirname / "temp" / "subfolder1" / BACKUP_FILENAME
    assert backup_file.is_file()
    backup_data = load_backup_file(backup_file)
    assert backup_data["sub1.pdf"][_kMDItemUserTags] == [["test", 0]]
    assert backup_data.get("sub1.txt") is None


def test_cli_order(test_dir):
    """Test order CLI options are executed

    Order of execution should be:
    restore, wipe, copyfrom, clear, set, append, remove, mirror, get, list, backup
    """

    dirname = pathlib.Path(test_dir)
    test_file = dirname / "test_file.txt"
    test_file.touch()
    test_file.write_text("test")

    md = OSXMetaData(test_file)
    md.tags = [Tag("test", 0)]
    md.authors = ["John Doe", "Jane Doe"]
    md.wherefroms = ["http://www.apple.com"]
    md.downloadeddate = [datetime.datetime(2019, 1, 1, 0, 0, 0)]
    md.findercomment = "Hello World"
    snooze(LONG_SNOOZE)

    runner = CliRunner()

    # first, create backup file for --restore
    runner.invoke(cli, ["--backup", test_file.as_posix()])

    # wipe the data
    runner.invoke(cli, ["--wipe", test_file.as_posix()])
    snooze(LONG_SNOOZE)

    # restore the data and check order of operations
    result = runner.invoke(
        cli,
        [
            "--get",
            "comment",
            "--set",
            "authors",
            "John Smith",
            "--restore",
            "--set",
            "title",
            "Test Title",
            "--clear",
            "title",
            "--append",
            "tags",
            "test2",
            "--set",
            "comment",
            "foo",
            "--remove",
            "authors",
            "Jane Doe",
            "--append",
            "authors",
            "Jane Smith",
            "--mirror",
            "comment",
            "findercomment",
            test_file.as_posix(),
        ],
    )
    output = parse_cli_output(result.output)
    assert output["comment"] == "Hello World"

    snooze(LONG_SNOOZE)
    md = OSXMetaData(test_file)
    assert md.authors == ["John Smith", "Jane Smith"]
    assert md.findercomment == "Hello World"
    assert md.tags == [Tag("test", 0), Tag("test2", 0)]



"""Test osxmetadata asdict and to_json method"""

import datetime
import json

from osxmetadata import ASDICT_ATTRIBUTES, OSXMetaData

from .conftest import snooze


def test_asdict(test_file):
    """Test asdict method"""
    md = OSXMetaData(test_file.name)
    md.authors = ["Jane Smith"]
    snooze()
    asdict = md.asdict()
    assert len(asdict) == len(ASDICT_ATTRIBUTES)
    assert asdict["kMDItemAuthors"] == ["Jane Smith"]


def test_asdict_subset(test_file):
    """Test asdict method with subset of attributes"""
    md = OSXMetaData(test_file.name)
    md.authors = ["Jane Smith"]
    snooze()
    asdict = md.asdict(attributes={"kMDItemAuthors"})
    assert len(asdict) == 1
    assert asdict["kMDItemAuthors"] == ["Jane Smith"]


def test_to_json(test_file):
    """Test to_json method"""
    md = OSXMetaData(test_file.name)
    md.authors = ["Jane Smith"]
    md.duedate = datetime.datetime(2022, 10, 1)
    snooze()
    json_str = md.to_json()
    json_data = json.loads(json_str)
    assert json_data["kMDItemAuthors"] == ["Jane Smith"]
    assert json_data["kMDItemDueDate"] == datetime.datetime(2022, 10, 1).isoformat()



"""Test Finder tags (_KMDItemUserTags) on a file."""

import pytest
from osxmetadata import *

from .conftest import snooze


def test_finder_tags(test_file):
    """Test Finder tags on a file."""

    md = OSXMetaData(test_file.name)
    assert not md.tags
    tag_values = [Tag("foo", FINDER_COLOR_NONE), Tag("bar", FINDER_COLOR_RED)]
    md.tags = tag_values
    snooze()
    assert sorted(md.tags) == sorted(tag_values)

    # test that tag names are being set correctly so NSURL can read them
    assert sorted(md.NSURLTagNamesKey) == ["bar", "foo"]

    # test that we can set tags to None
    md.tags = None
    snooze()
    assert not md.tags

    # test that we can set tags to empty list
    md.tags = []
    snooze()
    assert not md.tags


def test_finder_tags_get_set(test_file):
    """Test Finder tags on a file with get/set _kMDItemUserTags."""

    md = OSXMetaData(test_file.name)
    tag_values = [Tag("foo", FINDER_COLOR_NONE), Tag("bar", FINDER_COLOR_RED)]
    md.set(_kMDItemUserTags, tag_values)
    snooze()
    assert sorted(md.get(_kMDItemUserTags)) == sorted(tag_values)

    # test that tag names are being set correctly so NSURL can read them
    assert sorted(md.NSURLTagNamesKey) == ["bar", "foo"]

    # test that we can set tags to None
    md.set(_kMDItemUserTags, None)
    snooze()
    assert not md.get(_kMDItemUserTags)

    # test that we can set tags to empty list
    md.set(_kMDItemUserTags, [])
    snooze()
    assert not md.get(_kMDItemUserTags)



"""Test osxmetadata NSURL metadata tags"""

import pathlib

import pytest

from osxmetadata import OSXMetaData
from osxmetadata.attribute_data import (
    NSURL_RESOURCE_KEY_DATA,
)

from .conftest import snooze


@pytest.mark.parametrize("attribute_name", NSURL_RESOURCE_KEY_DATA.keys())
def test_nsurl_attributes_all(attribute_name, test_file):
    """Test that all NSURL attributes can be accessed without error"""

    md = OSXMetaData(test_file.name)
    md.get(attribute_name)


def test_nsurl_attribute_NSURLNameKey(test_file):
    """Test that NSURLNameKey can be read"""

    md = OSXMetaData(test_file.name)
    assert md.get("NSURLNameKey") == pathlib.Path(test_file.name).name


def test_nsurl_attribute_NSURLIsRegularFileKey(test_file):
    """Test that NSURLIsRegularFileKey can be read"""

    md = OSXMetaData(test_file.name)
    assert md.get("NSURLIsRegularFileKey") is True


def test_nsurl_attribute_NSURLTagNamesKey(test_file):
    """Test that NSURLTagNamesKey can be read and written"""

    md = OSXMetaData(test_file.name)
    assert md.get("NSURLTagNamesKey") is None
    md["NSURLTagNamesKey"] = ["a", "b"]
    snooze()
    assert md.NSURLTagNamesKey == ["a", "b"]
    md.NSURLTagNamesKey = []
    assert not md.NSURLTagNamesKey



"""Get and set macOS metadata keys for NSURL objects, for example, tags and labels"""

import typing as t

import CoreFoundation
import objc
from Foundation import NSURL


__all__ = ["get_nsurl_metadata", "set_nsurl_metadata"]


def get_nsurl_metadata(url: NSURL, key: str) -> t.Any:
    """Get metadata attribute value

    Args:
        key: metadata attribute name, e.g. NSURLTagNamesKey
    """
    success, value, error = url.getResourceValue_forKey_error_(None, key, None)
    if not success:
        raise ValueError(f"Error getting metadata for {key}: {error}")
    return value


def set_nsurl_metadata(url: NSURL, key: str, value: t.Any):
    """Set metadata attribute value

    Args:
        key: metadata attribute name, e.g. NSURLTagNamesKey
        value: value to set attribute to; must match the type expected by the attribute (e.g. str or list)
    """
    with objc.autorelease_pool():
        if isinstance(value, list):
            value = CoreFoundation.CFArrayCreate(
                None, value, len(value), CoreFoundation.kCFTypeArrayCallBacks
            )
        kv = CoreFoundation.NSDictionary.dictionaryWithObject_forKey_(value, key)
        success, error = url.setResourceValues_error_(kv, None)
        if not success:
            raise OSError(f"Error setting tag {key} = {value} on {url}: {error}")
        return True



""" Builds the attributes table in markdown format for README.md """

import osxmetadata

print("| Constant | Short Name | Long Constant | Description |")
print("|---------------|----------|---------|-----------|")
for attribute in sorted(
    [
        attribute_name
        for attribute_name in {
            osxmetadata.ATTRIBUTES[attr].name for attr in sorted(osxmetadata.ATTRIBUTES)
        }
    ]
):
    attr = osxmetadata.ATTRIBUTES[attribute]
    help_ = attr.api_help if attr.api_help is not None else attr.help
    print(f"|{attr.short_constant}|{attr.name}|{attr.constant}|{help_}|")



"""Load attribute metadata from JSON files"""

import datetime
import json
import pathlib
import typing as t

import Foundation

MDITEM_METADATA_FILES = [
    "audio_attributes.json",
    "common_attributes.json",
    "filesystem_attributes.json",
    "image_attributes.json",
    "video_attributes.json",
]

NSURL_RESOURCE_KEY_FILES = ["nsurl_resource_keys.json"]

MDIMPORTER_CONSTANT_FILES = ["mdimporter_constants.json"]


def load_mditem_attribute_data(files) -> t.Dict:
    """Load attribute metadata from JSON files"""
    # TODO: would be better to use importlib.abc.ResourceReader but I can't find a single example of how to do this
    parent = pathlib.Path(__file__).parent
    data = {}
    for filename in files:
        with open(parent / filename, "r") as f:
            file_data = json.load(f)
            for item in file_data:
                data[item["name"]] = item

    # add python types (e.g. CFString -> str) and help text types
    # the types used in the JSON files are the CF types with these counts
    # {'CFArray of CFStrings': 26, 'CFString': 61, 'CFNumber': 35, 'CFBoolean': 8, 'CFDate': 8}
    for key, value in data.items():
        if value["type"] == "CFString":
            data[key]["python_type"] = "str"
            data[key]["help_type"] = "string"
        elif value["type"] == "CFNumber":
            data[key]["python_type"] = "float"
            data[key]["help_type"] = "number"
        elif value["type"] == "CFBoolean":
            data[key]["python_type"] = "bool"
            data[key]["help_type"] = "boolean"
        elif value["type"] == "CFDate":
            data[key]["python_type"] = "datetime.datetime"
            data[key]["help_type"] = "date/time"
        elif value["type"] == "CFArray of CFStrings":
            data[key]["python_type"] = "list"
            data[key]["help_type"] = "list of strings"
        else:
            raise ValueError(f"Unknown type {value['type']}")

    # add the attribute short_name which is lowercase of the part after kMDItem
    for key, value in data.items():
        if value["name"].startswith("kMDItem"):
            data[key]["short_name"] = value["name"][7:].lower()
        else:
            data[key]["short_name"] = value["name"].lower()

    # add the xattr_constant which is com.apple.metadata:MDItemName
    for key, value in data.items():
        data[key]["xattr_constant"] = f"com.apple.metadata:{value['name']}"

    return data


def load_nsurl_resource_key_data() -> t.Dict:
    """Load NSRL resource key metadata from JSON files"""
    # TODO: would be better to use importlib.abc.ResourceReader but I can't find a single example of how to do this
    parent = pathlib.Path(__file__).parent
    data = {}
    for filename in NSURL_RESOURCE_KEY_FILES:
        with open(parent / filename, "r") as f:
            file_data = json.load(f)
            for item in file_data:
                # Some resource keys only available on certain versions of macOS
                try:
                    getattr(Foundation, item["name"])
                except AttributeError:
                    # Not available on this version of macOS
                    continue
                data[item["name"]] = item
    return data


def load_mdimporter_constant_data() -> t.Dict:
    """Load MDImporter constants from JSON files"""
    parent = pathlib.Path(__file__).parent
    data = {}
    for filename in MDIMPORTER_CONSTANT_FILES:
        with open(parent / filename, "r") as f:
            file_data = json.load(f)
            for item in file_data:
                data[item["name"]] = item
    return data


# all attribute data
MDITEM_ATTRIBUTE_DATA = load_mditem_attribute_data(MDITEM_METADATA_FILES)


# Add kMDItemDownloadedDate which isn't in the normal MDItem reference but is
# referenced in the MDImporter reference here:
# https://developer.apple.com/documentation/coreservices/kmditemdownloadeddate?language=objc
MDITEM_ATTRIBUTE_DATA["kMDItemDownloadedDate"] = {
    "name": "kMDItemDownloadedDate",
    "description": "Date the item was downloaded.",
    "version": "10.7",
    "type": "CFArray of CFDates",
    "python_type": "list[datetime.datetime]",
    "help_type": "list of date/time",
    "short_name": "downloadeddate",
    "xattr_constant": "com.apple.metadata:kMDItemDownloadedDate",
}

# specific types of attribute data
MDITEM_ATTRIBUTE_AUDIO = load_mditem_attribute_data(["audio_attributes.json"])
MDITEM_ATTRIBUTE_COMMON = load_mditem_attribute_data(["common_attributes.json"])
MDITEM_ATTRIBUTE_FILESYSTEM = load_mditem_attribute_data(["filesystem_attributes.json"])
MDITEM_ATTRIBUTE_IMAGE = load_mditem_attribute_data(["image_attributes.json"])
MDITEM_ATTRIBUTE_VIDEO = load_mditem_attribute_data(["video_attributes.json"])

# short names for accessing via OSXMetaData().short_name
MDITEM_ATTRIBUTE_SHORT_NAMES = {
    item["short_name"]: item["name"] for item in MDITEM_ATTRIBUTE_DATA.values()
}

# Some MDItem attributes are read-only
MDITEM_ATTRIBUTE_READ_ONLY = {
    "kMDItemAttributeChangeDate",
    "kMDItemContentType",
    "kMDItemKind",
    "kMDItemLastUsedDate",
    "kMDItemSecurityMethod",
    "kMDItemTextContent",
    "kMDItemFSHasCustomIcon",
    "kMDItemFSIsStationery",
    "kMDItemDisplayName",
    "kMDItemFSContentChangeDate",
    "kMDItemFSCreationDate",
    "kMDItemFSInvisible",
    "kMDItemFSIsExtensionHidden",
    "kMDItemFSLabel",
    "kMDItemFSName",
    "kMDItemFSNodeCount",
    "kMDItemFSOwnerGroupID",
    "kMDItemFSOwnerUserID",
    "kMDItemFSSize",
    "kMDItemPath",
}

NSURL_RESOURCE_KEY_DATA = load_nsurl_resource_key_data()

MDIMPORTER_ATTRIBUTE_DATA = load_mdimporter_constant_data()


__all__ = [
    "MDITEM_ATTRIBUTE_DATA",
    "MDITEM_ATTRIBUTE_READ_ONLY",
    "MDITEM_ATTRIBUTE_SHORT_NAMES",
    "MDITEM_ATTRIBUTE_AUDIO",
    "MDITEM_ATTRIBUTE_COMMON",
    "MDITEM_ATTRIBUTE_FILESYSTEM",
    "MDITEM_ATTRIBUTE_IMAGE",
    "MDITEM_ATTRIBUTE_VIDEO",
    "NSURL_RESOURCE_KEY_DATA",
    "MDIMPORTER_ATTRIBUTE_DATA",
]



""" OSXMetaData class to read and write various Mac OS X metadata 
    such as tags/keywords and Finder comments from files """

import base64
import datetime
import json
import pathlib
import typing as t

import CoreServices
import xattr
from Foundation import NSURL

from ._version import __version__
from .attribute_data import (
    MDIMPORTER_ATTRIBUTE_DATA,
    MDITEM_ATTRIBUTE_DATA,
    MDITEM_ATTRIBUTE_SHORT_NAMES,
    NSURL_RESOURCE_KEY_DATA,
)
from .finder_comment import kMDItemFinderComment, set_or_remove_finder_comment
from .finder_info import (
    _kFinderColor,
    _kFinderInfo,
    _kFinderStationeryPad,
    get_finderinfo_bytes,
    get_finderinfo_color,
    get_finderinfo_stationerypad,
    set_finderinfo_bytes,
    set_finderinfo_color,
    set_finderinfo_stationerypad,
)
from .finder_tags import _kMDItemUserTags, get_finder_tags, set_finder_tags
from .mditem import (
    MDItemValueType,
    get_mditem_metadata,
    set_or_remove_mditem_metadata,
    MDItemSetAttribute,
)
from .nsurl_metadata import get_nsurl_metadata, set_nsurl_metadata

ALL_ATTRIBUTES = {
    "finderinfo",
    "tags",
    *list(MDITEM_ATTRIBUTE_DATA.keys()),
    *list(MDITEM_ATTRIBUTE_SHORT_NAMES.keys()),
    *list(NSURL_RESOURCE_KEY_DATA.keys()),
    *list(MDIMPORTER_ATTRIBUTE_DATA.keys()),
    _kFinderColor,
    _kFinderInfo,
    _kFinderStationeryPad,
    _kMDItemUserTags,
}

# Subset of attributes returned by asdict() and to_json() methods
ASDICT_ATTRIBUTES = {
    *list(MDITEM_ATTRIBUTE_DATA.keys()),
    *list(MDIMPORTER_ATTRIBUTE_DATA.keys()),
    _kFinderStationeryPad,
    _kFinderColor,
    _kMDItemUserTags,
}


class OSXMetaDataAttributeError(Exception):
    """Raised when an attribute is not supported or attempting to set read-only attribute"""

    pass


class OSXMetaData:
    """Create an OSXMetaData object to access file metadata"""

    def __init__(self, fname: str):
        """Create an OSXMetaData object to access file metadata
        fname: filename to operate on
        """
        self._fname = pathlib.Path(fname)
        if not self._fname.exists():
            raise FileNotFoundError(f"file does not exist: {fname}")

        self._posix_path = self._fname.resolve().as_posix()

        # Create MDItemRef, NSURL, and xattr objects
        # MDItemRef is used for most attributes
        # NSURL and xattr are required for certain attributes like Finder tags
        # Because many of the getter/setter functions require some combination of MDItemRef, NSURL, and xattr,
        # they are created here and kept for the life of the object so that they don't have to be
        # recreated for each attribute
        # This does mean that if the file is moved or renamed, the object will still be pointing to the old file
        # thus you should not rename or move a file while using an OSXMetaData object
        self._mditem: CoreServices.MDItemRef = CoreServices.MDItemCreate(
            None, self._posix_path
        )
        if not self._mditem:
            raise OSError(f"Unable to create MDItem for file: {fname}")
        self._url = NSURL.fileURLWithPath_(self._posix_path)
        self._xattr = xattr.xattr(self._posix_path)

        # Required so __setattr__ gets handled correctly during __init__
        self.__init = True

    def get(self, attribute: str) -> MDItemValueType:
        """Get metadata attribute value
        attribute: metadata attribute name
        """
        return self.__getattr__(attribute)

    def set(self, attribute: str, value: MDItemValueType):
        """Set metadata attribute value

        Args:
            attribute: metadata attribute name
            value: value to set attribute to; must match the type expected by the attribute (e.g. str or list)
        """
        self.__setattr__(attribute, value)

    def get_xattr(
        self, key: str, decode: t.Callable[[t.ByteString], t.Any] = None
    ) -> t.Any:
        """Get xattr value

        Args:
            key: xattr name
            decode: optional Callable to decode value before returning
        """
        xattr = self._xattr[key]
        if decode:
            xattr = decode(xattr)
        return xattr

    def set_xattr(
        self, key: str, value: t.Any, encode: t.Callable[[t.ByteString], t.Any] = None
    ):
        """Set xattr value

        Args:
            key: xattr name
            encode: optional Callable to encode value before setting
        """
        if encode:
            value = encode(value)
        self._xattr[key] = value

    def remove_xattr(self, key: str):
        """Remove xattr

        Args:
            key: xattr name
        """
        self._xattr.remove(key)

    def asdict(self, attributes: t.Set[str] = ASDICT_ATTRIBUTES) -> t.Dict[str, t.Any]:
        """Return all MDItem metadata (or a subset defined by attributes) as a dict

        Args:
            attributes: set of attributes to include in dict

        Returns:
            dict of metadata
        """
        return {key: getattr(self, key) for key in attributes}

    def to_json(
        self, attributes: t.Set[str] = ASDICT_ATTRIBUTES, indent: int = 4
    ) -> str:
        """Return all MDItem metadata (or a subset defined by attributes) as a JSON string

        Args:
            attributes: set of attributes to include in JSON
            indent: indent level for JSON output

        Returns:
            JSON string

        Notes:
            datetime objects are converted to ISO 8601 format
            binary objects are converted to base64 encoded strings
            the resulting JSON will include 3 additional keys: _version, _filepath, and _filename;
            these are expected by the CLI backup/restore commands
        """

        dict_data = self.asdict(attributes)

        # add fields that backup/restore expects
        dict_data.update(
            {
                "_version": __version__,
                "_filepath": self._posix_path,
                "_filename": self._fname.name,
            }
        )

        for key, value in dict_data.items():
            if isinstance(value, datetime.datetime):
                dict_data[key] = value.isoformat()
            elif isinstance(value, (list, tuple)):
                if value and isinstance(value[0], datetime.datetime):
                    dict_data[key] = [v.isoformat() for v in value]
            elif isinstance(value, bytes):
                dict_data[key] = base64.b64encode(value).decode("ascii")

        return json.dumps(dict_data, indent=indent)

    def get_mditem_attribute_value(self, attribute: str) -> t.Any:
        """Get the raw MDItem attribute value without any type conversion.

        Args:
            attribute: metadata attribute name

        Returns:
            raw MDItem attribute value as returned by CoreServices.MDItemCopyAttribute()

        Note: This is a low level function that you probably don't need to use,
        but may be useful in some cases. You should probably use the get() method instead.
        """
        return CoreServices.MDItemCopyAttribute(self._mditem, attribute)

    def set_mditem_attribute_value(self, attribute: str, value: t.Any) -> bool:
        """Set the raw MDItem attribute value without any type conversion.

        Args:
            attribute: metadata attribute name
            value: value to set attribute to

        Returns: True if successful otherwise False

        Note: This is a low level function that you probably don't need to use,
        but may be useful in some cases. You should probably use the set() method instead.
        """
        return MDItemSetAttribute(self._mditem, attribute, value)

    @property
    def path(self) -> str:
        """Return path to file"""
        return self._posix_path

    def __getattr__(self, attribute: str) -> MDItemValueType:
        """Get metadata attribute value

        Args:
            attribute: metadata attribute name
        """
        if attribute in ["tags", _kMDItemUserTags]:
            return get_finder_tags(self._xattr)
        elif attribute in MDITEM_ATTRIBUTE_SHORT_NAMES:
            # handle dynamic properties like self.keywords and self.comments
            return get_mditem_metadata(
                self._mditem, MDITEM_ATTRIBUTE_SHORT_NAMES[attribute]
            )
        elif (
            attribute in MDITEM_ATTRIBUTE_DATA or attribute in MDIMPORTER_ATTRIBUTE_DATA
        ):
            return get_mditem_metadata(self._mditem, attribute)
        elif attribute in NSURL_RESOURCE_KEY_DATA:
            return get_nsurl_metadata(self._url, attribute)
        elif attribute in ["finderinfo", _kFinderInfo]:
            return get_finderinfo_bytes(self._xattr)
        elif attribute == _kFinderStationeryPad:
            return get_finderinfo_stationerypad(self._xattr)
        elif attribute == _kFinderColor:
            return get_finderinfo_color(self._xattr)
        else:
            raise AttributeError(f"Invalid attribute: {attribute}")

    def __setattr__(self, attribute: str, value: t.Any):
        """set metadata attribute value

        Args:
            attribute: metadata attribute name
            value: value to set
        """
        try:
            if not self.__init:
                # during __init__ we don't want to call __setattr__ as it will
                # cause an infinite loop
                return super().__setattr__(attribute, value)
            if attribute in ["findercomment", kMDItemFinderComment]:
                # finder comment cannot be set using MDItemSetAttribute
                set_or_remove_finder_comment(self._url, self._xattr, value)
            elif attribute in ["tags", _kMDItemUserTags]:
                # handle Finder tags
                set_finder_tags(self._url, value)
            elif attribute in MDITEM_ATTRIBUTE_SHORT_NAMES:
                # handle dynamic properties like self.keywords and self.comments
                attribute_name = MDITEM_ATTRIBUTE_SHORT_NAMES[attribute]
                set_or_remove_mditem_metadata(self._mditem, attribute_name, value)
            elif attribute in MDITEM_ATTRIBUTE_DATA:
                set_or_remove_mditem_metadata(self._mditem, attribute, value)
            elif attribute in NSURL_RESOURCE_KEY_DATA:
                set_nsurl_metadata(self._url, attribute, value)
            elif attribute in ["finderinfo", _kFinderInfo]:
                set_finderinfo_bytes(self._xattr, value)
            elif attribute == _kFinderStationeryPad:
                set_finderinfo_stationerypad(self._xattr, bool(value))
            elif attribute == _kFinderColor:
                set_finderinfo_color(self._xattr, value)
            elif attribute in ALL_ATTRIBUTES:
                raise OSXMetaDataAttributeError(f"Attribute {attribute} is read-only")
            else:
                raise OSXMetaDataAttributeError(f"Invalid attribute: {attribute}")
        except (KeyError, AttributeError):
            super().__setattr__(attribute, value)
        except OSXMetaDataAttributeError as e:
            raise AttributeError(e) from e

    def __getitem__(self, key: str) -> MDItemValueType:
        """Get metadata attribute value

        Args:
            key: metadata attribute name
        """
        if key == _kMDItemUserTags:
            return get_finder_tags(self._xattr)
        elif key in MDITEM_ATTRIBUTE_DATA:
            return get_mditem_metadata(self._mditem, key)
        elif key in NSURL_RESOURCE_KEY_DATA:
            return get_nsurl_metadata(self._url, key)
        else:
            raise KeyError(f"Invalid key: {key}")

    def __setitem__(self, key: str, value: t.Any):
        """set metadata attribute value

        Args:
            key: metadata attribute name
            value: value to set
        """
        if key == _kMDItemUserTags:
            set_finder_tags(self._xattr, value)
        elif key == kMDItemFinderComment:
            set_or_remove_finder_comment(self._url, self._xattr, value)
        elif key in MDITEM_ATTRIBUTE_DATA:
            set_or_remove_mditem_metadata(self._mditem, key, value)
        elif key in NSURL_RESOURCE_KEY_DATA:
            set_nsurl_metadata(self._url, key, value)
        elif key in ALL_ATTRIBUTES:
            raise KeyError(f"Attribute {key} is read-only")
        else:
            raise KeyError(f"Invalid key: {key}")
